"""
.. autoclass:: ContextT

.. autoclass:: DeviceT
"""

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol


class DeviceT(Protocol):
    """
    Abstract type providing an interface like :class:`pyopencl.Device`.
    """

    @property
    def name(self) -> str:
        pass


class ContextT(Protocol):
    """
    Abstract type providing an interface like :class:`pyopencl.Context`.
    """

    @property
    def devices(self) -> Sequence[DeviceT]:
        pass


# {{{ make_fake_cl_context


@dataclass(frozen=True, repr=True, eq=True)
class FakeDevice:
    name: str


@dataclass(frozen=True, repr=True, eq=True)
class FakeContext:
    devices: tuple[FakeDevice, ...]


def make_fake_cl_context(devices: str | Sequence[str]) -> FakeContext:
    if isinstance(devices, str):
        devices = (devices,)

    return FakeContext(tuple(FakeDevice(dev_name) for dev_name in devices))


# }}}



"""
.. currentmodule:: feinsum.einsum

.. autoclass:: BatchedEinsum
.. autoclass:: EinsumAxisAccess
.. autoclass:: FreeAxis
.. autoclass:: SummationAxis
.. autoclass:: Argument
.. autoclass:: Array
.. autoclass:: EinsumOperand
.. autoclass:: IntermediateResult
.. autoclass:: ContractionSchedule
.. autoclass:: SizeParam


Helper routines
^^^^^^^^^^^^^^^

.. autofunction:: get_trivial_contraction_schedule
.. autofunction:: get_opt_einsum_contraction_schedule
"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import Any, Self, cast

import numpy as np
from immutables import Map
from pytools import UniqueNameGenerator, memoize_method

IntegralT = int | np.integer
INT_CLASSES = (int, np.integer)


@dataclass(frozen=True)
class SizeParam:
    name: str

    def __truediv__(self, other: IntegralT | SizeParam) -> IntegralT:
        # only present to help writing easier param-getters in the tuner
        # implementations.
        return NotImplemented

    __rtruediv__ = __truediv__


ShapeComponentT = IntegralT | SizeParam
ShapeT = tuple[ShapeComponentT, ...]


@dataclass(frozen=True, eq=True, repr=True)
class Array:
    name: str
    shape: ShapeT
    dtype: np.dtype[Any]

    @property
    def ndim(self) -> int:
        return len(self.shape)

    def copy(
        self,
        *,
        name: str | None = None,
        shape: ShapeT | None = None,
        dtype: np.dtype[Any] | None = None,
    ) -> Self:
        from dataclasses import replace

        return replace(
            self,
            name=self.name if name is None else name,
            shape=self.shape if shape is None else shape,
            dtype=self.dtype if dtype is None else dtype,
        )


@dataclass(frozen=True)
class EinsumAxisAccess:
    """
    Base class for axis access types in an einsum expression.
    """

    def __init__(self) -> None:
        if type(self) is EinsumAxisAccess:
            raise TypeError(
                "EinsumAxisAccess is abstract and cannot be instantiated directly"
            )


@dataclass(frozen=True)
class FreeAxis(EinsumAxisAccess):
    """
    Records the axis of an einsum argument over which contraction is not performed.

    .. attribute:: output_index

        Position of the corresponding index in the einsum's output.
    """

    output_index: int


@dataclass(frozen=True)
class SummationAxis(EinsumAxisAccess):
    """
    Records an index in an einsum expression over which reduction is performed.
    Sometimes also referred to as an axis with a corresponding "dummy index" in
    Ricci Calculus.

    .. attribute:: index

        An integer which is unique to a reduction index of an einsum.
    """

    index: int


@dataclass(frozen=True)
class BatchedEinsum:
    """
    A batched einsum expression.

    .. attribute:: output_indices
    .. attribute:: input_indices
    .. attribute:: args
    """

    out_idx_set: tuple[str, ...]
    in_idx_sets: tuple[tuple[str, ...], ...]
    args: tuple[tuple[Array, ...], ...]

    def __post_init__(self) -> None:
        from functools import reduce

        assert all(
            len(idx) == 1 and idx.islower() for idx in self.out_idx_set
        ), "Obtained invalid output index (RHS of ->)."
        assert all(
            all(len(idx) == 1 and idx.islower() for idx in in_idx_set)
            for in_idx_set in self.in_idx_sets
        ), "Obtained invalid input index (LHS of ->)."
        all_in_indices = reduce(
            frozenset.union,
            (frozenset(idx_set) for idx_set in self.in_idx_sets),
            cast("frozenset[str]", frozenset()),
        )
        assert (
            frozenset(self.out_idx_set) <= all_in_indices
        ), "Obtained an out index which is not present in the input indices."

        assert all(
            len(arg_row) == len(self.in_idx_sets) for arg_row in self.args
        ), "Mismatch in #operands between subscript expression and input arrays."
        assert all(
            all(
                arg.ndim == len(idx_set)
                for arg, idx_set in zip(arg_row, self.in_idx_sets, strict=True)
            )
            for arg_row in self.args
        ), "Dimensionality of input operands do no match the provided subscripts."

        _ = self.arg_to_dtype
        _ = self.arg_to_shape
        _ = self.index_to_dim_length
        assert (
            len(self.all_args) + len(self.all_indices) + len(self.all_size_params)
        ) == len(
            self.all_args | self.all_indices | {p.name for p in self.all_size_params}
        ), "Must use different names for arguments, indices, and size params."

    @cached_property
    def b(self) -> int:
        """
        Returns the number of batches in the batched einsum.
        """
        return len(self.args)

    @cached_property
    def n(self) -> int:
        """
        Return the number of operands of each einsum in the batched einsum.
        """
        return len(self.in_idx_sets)

    @cached_property
    def index_to_dim_length(self) -> Map[str, ShapeComponentT]:
        index_to_dim: dict[str, ShapeComponentT] = {}
        for arg_row in self.args:
            for arg, idx_set in zip(arg_row, self.in_idx_sets, strict=True):
                for axis_len, idx in zip(arg.shape, idx_set, strict=True):
                    assert (
                        index_to_dim.setdefault(idx, axis_len) == axis_len
                    ), "Shape mismatch for indices across the arguments."

        return Map(index_to_dim)

    @cached_property
    def shape(self) -> ShapeT:
        """
        Returns the shape of an output of the batched einsum.
        """
        return tuple(self.index_to_dim_length[idx] for idx in self.out_idx_set)

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @memoize_method
    def get_subscripts(self) -> str:
        """
        Returns the subscripts used in the building the *einsum* from it.
        """
        joined_input_induces = ["".join(idx_set) for idx_set in self.in_idx_sets]
        return f"{','.join(joined_input_induces)} -> {''.join(self.out_idx_set)}"

    @cached_property
    def arg_to_shape(self) -> Map[str, ShapeT]:
        result: dict[str, ShapeT] = {}
        for arg_row in self.args:
            for arg in arg_row:
                assert (
                    result.setdefault(arg.name, arg.shape) == arg.shape
                ), f"Inconsistent shapes for arg {arg.name}."
        return Map(result)

    @cached_property
    def arg_to_dtype(self) -> Map[str, np.dtype[Any]]:
        result: dict[str, np.dtype[Any]] = {}
        for arg_row in self.args:
            for arg in arg_row:
                assert (
                    result.setdefault(arg.name, arg.dtype) == arg.dtype
                ), f"Inconsistent dtypes for arg {arg.name}."
        return Map(result)

    @cached_property
    def index_to_access_descr(self) -> Map[str, EinsumAxisAccess]:
        result: dict[str, EinsumAxisAccess] = {}
        for i, idx in enumerate(self.out_idx_set):
            result[idx] = FreeAxis(i)

        i_redn = 0
        for idx_set in self.in_idx_sets:
            for idx in idx_set:
                if idx not in result:
                    result[idx] = SummationAxis(i_redn)
                    i_redn += 1
        return Map(result)

    @cached_property
    def sum_indices(self) -> tuple[str, ...]:
        all_sum_indices = {
            idx: access.index
            for idx, access in self.index_to_access_descr.items()
            if isinstance(access, SummationAxis)
        }
        return tuple(sorted(all_sum_indices, key=lambda x: all_sum_indices[x]))

    @cached_property
    def all_args(self) -> frozenset[str]:
        return frozenset(self.arg_to_shape)

    @cached_property
    def all_indices(self) -> frozenset[str]:
        return frozenset(self.index_to_dim_length)

    @cached_property
    def all_size_params(self) -> frozenset[SizeParam]:
        return frozenset(
            v for v in self.index_to_dim_length.values() if isinstance(v, SizeParam)
        )

    def copy(self, **kwargs: Any) -> BatchedEinsum:
        from dataclasses import replace

        return replace(self, **kwargs)


class Argument:
    """
    An abstract class denoting an argument to an einsum in
    :class:`ContractionSchedule`. See :attr:`ContractionSchedule.arguments`.
    """

    def __init__(self) -> None:
        if type(self) is Argument:
            raise TypeError(
                "Argument is abstract and cannot be instantiated directly."
            )


@dataclass(frozen=True)
class IntermediateResult(Argument):
    """
    An :class:`Argument` representing an intermediate result available during
    the current contraction.
    """

    name: str


@dataclass(frozen=True, eq=True, repr=True)
class EinsumOperand(Argument):
    """
    An :class:`Argument` representing the *ioperand*-th argument that was
    passed to the parent einsum whose :class:`ContractionSchedule` is being
    specified.
    """

    ioperand: int


@dataclass(frozen=True, eq=True, repr=True)
class ContractionSchedule:
    """
    Records the schedule in which contractions are to be performed in an einsum
    as a series of einsums with the i-th einsum having subscript
    ``subscript[i]`` operating on ``arguments[i]`` and writing its result to
    ``result_names[i]``.

    .. attribute:: result_names

        Names of the result generated by each step.

    .. attribute:: arguments

       A :class:`tuple` containing :class:`tuple` of :class:`str` for each
       contraction in the schedule.

    .. attribute:: nsteps
    """

    subscripts: tuple[str, ...]
    result_names: tuple[str, ...]
    arguments: tuple[tuple[Argument, ...], ...]

    def __post_init__(self) -> None:
        assert len(self.subscripts) == len(self.result_names) == len(self.arguments)

    @property
    def nsteps(self) -> int:
        """
        Returns the number of steps involved in scheduling the einsum.
        """
        return len(self.subscripts)

    def copy(self, **kwargs: Any) -> ContractionSchedule:
        from dataclasses import replace

        return replace(self, **kwargs)


def get_trivial_contraction_schedule(einsum: BatchedEinsum) -> ContractionSchedule:
    """
    Returns the :class:`ContractionSchedule` for *einsum* scheduled as a single
    contraction.
    """
    return ContractionSchedule(
        (einsum.get_subscripts(),),
        ("_fe_out",),
        (tuple(EinsumOperand(i) for i in range(einsum.n)),),
    )


def get_opt_einsum_contraction_schedule(
    expr: BatchedEinsum,
    **opt_einsum_kwargs: Any,
) -> ContractionSchedule:
    """
    Returns a :class:`ContractionSchedule` as computed by
    :func:`opt_einsum.contract_path`.

    :param opt_einsum_kwargs: kwargs to be passed to
        :func:`opt_einsum.contract_path`.

    .. note::

        The following defaults are populated in *opt_einsum_kwargs*, if left
        unspecified:

        - ``optimize="optimal"``
        - ``use_blas=False``
    """
    import opt_einsum

    long_dim_length = opt_einsum_kwargs.pop("long_dim_length", 1_000_000)

    if "optimize" not in opt_einsum_kwargs:
        opt_einsum_kwargs["optimize"] = "optimal"

    if "use_blas" not in opt_einsum_kwargs:
        opt_einsum_kwargs["use_blas"] = False

    _, path = opt_einsum.contract_path(
        expr.get_subscripts(),
        *[
            arg.copy(
                shape=tuple(
                    long_dim_length if isinstance(dim, SizeParam) else dim
                    for dim in arg.shape
                )
            )
            for arg in expr.args[0]
        ],
        **opt_einsum_kwargs,
    )

    current_args: list[Argument] = [
        EinsumOperand(i) for i in range(path.input_subscripts.count(",") + 1)
    ]
    vng = UniqueNameGenerator()

    subscripts: list[str] = []
    result_names: list[str] = []
    arguments: list[tuple[Argument, ...]] = []
    for contraction in path.contraction_list:
        arg_indices, _, subscript, _, _ = contraction
        arguments.append(tuple(current_args[idx] for idx in arg_indices))
        subscripts.append(subscript)
        result_names.append(vng("_fe_tmp"))
        current_args = [
            arg for idx, arg in enumerate(current_args) if idx not in arg_indices
        ] + [IntermediateResult(result_names[-1])]

    assert len(current_args) == 1
    result_names[-1] = vng("_fe_out")

    return ContractionSchedule(
        tuple(subscripts), tuple(result_names), tuple(arguments)
    )



__copyright__ = """Copyright (C) 2022 Kaushik Kulkarni"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from pyopencl.tools import (  # noqa
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)

import feinsum as f
import loopy as lp
import numpy as np
import feinsum.loopy_utils as lp_utils


def test_extract_subexpr_of_associative_op_as_subst(ctx_factory):
    nface = 4
    nvoldofs = 35
    nfacedofs = 15
    face_mass = f.batched_einsum(
        "ef, fij, fej -> ei",
        [
            [
                f.array("J", ("E", nface)),
                f.array("R", (nface, nvoldofs, nfacedofs)),
                f.array(f"v{i}", (nface, "E", nfacedofs)),
            ]
            for i in range(4)
        ],
    )
    t_unit = f.generate_loopy(face_mass)
    sigma = f.match_t_unit_to_einsum(t_unit, face_mass)

    output_names = ["_fe_out"] + [f"_fe_out_{i}" for i in range(3)]

    from loopy.symbolic import get_dependencies
    from pymbolic import variables

    knl = t_unit.default_entrypoint
    iname_f, iname_e, iname_j = sigma["f"], sigma["e"], sigma["j"]

    for i in range(4):
        knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
            knl,
            within=f"writes:{sigma[output_names[i]]}",
            subst_name=f"subst_{i}",
            arguments=variables(f"{iname_f} {iname_e} {iname_j}"),
            terms_filter=lambda x: (get_dependencies(x) & knl.all_inames())
            <= {iname_f, iname_e, iname_j},
        )

    t_unit = t_unit.with_kernel(knl)

    for i in range(4):
        t_unit = lp.precompute(
            t_unit,
            f"subst_{i}",
            sweep_inames=[iname_f, iname_j],
            precompute_outer_inames=frozenset({iname_e}),
            default_tag=None,
            temporary_address_space=lp.AddressSpace.PRIVATE,
        )

    opt_einsum_t_unit = f.generate_loopy_with_opt_einsum_schedule(face_mass)

    assert (lp.get_op_map(t_unit, subgroup_size=1).eval_and_sum({"E": 1})) == (
        lp.get_op_map(opt_einsum_t_unit, subgroup_size=1).eval_and_sum({"E": 1})
    )


def test_hoist_reduction_invariant_terms(ctx_factory):
    from loopy.symbolic import Reduction
    from pymbolic import variables

    cl_ctx = ctx_factory()
    nel = 1
    ndim = 3
    ndofs = 35
    expr = f.einsum(
        "xre, rij, ej->xei",
        f.array("J", (ndim, ndim, nel)),
        f.array("R", (ndim, ndofs, ndofs)),
        f.array("u", (nel, ndofs)),
    )
    t_unit = f.generate_loopy(expr)
    sigma = f.match_t_unit_to_einsum(t_unit, expr)

    # {{{ hoist the "j" redn-loop over "x" loop

    e, i, j, r = sigma["e"], sigma["i"], sigma["j"], sigma["r"]
    knl = t_unit.default_entrypoint
    knl = lp.split_reduction_inward(knl, j)

    knl = lp_utils.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, j)
    knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
        knl,
        within=None,
        subst_name="grad_without_jacobi_subst",
        arguments=variables(f"{r} {i} {e}"),
        terms_filter=lambda x: isinstance(x, Reduction),
    )

    hoisted_t_unit = t_unit.with_kernel(knl)

    hoisted_t_unit = lp.precompute(
        hoisted_t_unit,
        "grad_without_jacobi_subst",
        sweep_inames=[r, i],
        precompute_outer_inames=frozenset({e}),
    )

    # }}}

    trivial_ops = lp.get_op_map(t_unit, subgroup_size=1).eval_and_sum()
    opt_ops = lp.get_op_map(hoisted_t_unit, subgroup_size=1).eval_and_sum()
    assert trivial_ops > 4 * opt_ops
    lp.auto_test_vs_ref(t_unit, cl_ctx, hoisted_t_unit)


def test_wave_grad_transform_knowledge_transfer(ctx_factory):
    from testlib import transform_3d_p4_grad

    cl_ctx = ctx_factory()

    t_unit = lp.make_kernel(
        "{[iel_1, idof_1, jdof_1, r_1, x_1]:"
        " 0<=iel_1<Nel and 0<=idof_1,jdof_1<35 and 0<=r_1,x_1<3}",
        """
        jac_subst(_0, _1, _2) := J[_0, _1, _2]
        D_subst(_0, _1, _2) := D[_0, _1, _2]
        u_subst(_0, _1) := u[_0, _1]

        grad_out[x_1, iel_1, idof_1] = sum([jdof_1, r_1], \
                                           jac_subst(x_1, iel_1, r_1)*D_subst(r_1, idof_1, jdof_1)*u_subst(iel_1, jdof_1))
        """,  # noqa: E501
    )
    t_unit = lp.add_dtypes(
        t_unit,
        {
            arg.name: np.float64
            for arg in t_unit.default_entrypoint.args
            if arg.is_input and arg.name != "Nel"
        },
    )
    ref_t_unit = t_unit
    t_unit = transform_3d_p4_grad(t_unit, "writes:grad_out")
    lp.auto_test_vs_ref(ref_t_unit, cl_ctx, t_unit, parameters={"Nel": 100})


def test_einsum_matching():
    t_unit = lp.make_kernel(
        "{[iel_2, idof_2, ifacedof, iface]:"
        " 0<=iel_2<10000 and 0<=idof_2<35 and 0<=ifacedof<15 and 0<=iface<4}",
        """
        flux_subst_0(_0, _1, _2) := F_0[_0, _1, _2]
        flux_subst_1(_0, _1, _2) := F_1[_0, _1, _2]
        flux_subst_2(_0, _1, _2) := F_2[_0, _1, _2]
        flux_subst_3(_0, _1, _2) := F_3[_0, _1, _2]
        lift_subst(_0, _1, _2) := Rlift[_0, _1, _2]
        face_jac_subst(_0, _1) := Jface[_0, _1]

        lift_0[iel_2, idof_2] = sum([iface, ifacedof],
                                    face_jac_subst(iel_2, iface)*lift_subst(iface, idof_2, ifacedof)*flux_subst_0(iface, iel_2, ifacedof))
        lift_1[iel_2, idof_2] = sum([iface, ifacedof],
                                    face_jac_subst(iel_2, iface)*lift_subst(iface, idof_2, ifacedof)*flux_subst_1(iface, iel_2, ifacedof))
        lift_2[iel_2, idof_2] = sum([iface, ifacedof],
                                    face_jac_subst(iel_2, iface)*lift_subst(iface, idof_2, ifacedof)*flux_subst_2(iface, iel_2, ifacedof))
        lift_3[iel_2, idof_2] = sum([iface, ifacedof],
                                    face_jac_subst(iel_2, iface)*lift_subst(iface, idof_2, ifacedof)*flux_subst_3(iface, iel_2, ifacedof))
        """,  # noqa: E501
    )

    t_unit = lp.add_dtypes(
        t_unit,
        {
            arg.name: np.float64
            for arg in t_unit.default_entrypoint.args
            if arg.is_input
        },
    )

    nvoldofs = 35
    nfacedofs = 15
    nface = 4

    ref_einsum = f.batched_einsum(
        "ef, fij, fej -> ei",
        [
            [
                f.array("J", ("E", nface)),
                f.array("R", (nface, nvoldofs, nfacedofs)),
                f.array(f"v{i}", (nface, "E", nfacedofs)),
            ]
            for i in range(4)
        ],
    )

    inferred_einsum, _ = f.get_a_matched_einsum(t_unit)
    assert f.canonicalize_einsum(inferred_einsum) == f.canonicalize_einsum(
        ref_einsum
    )


def test_sum_redn_algebraic_transforms(ctx_factory):
    from loopy.symbolic import Reduction
    from pymbolic import variables

    t_unit = lp.make_kernel(
        "{[e,i,j,x,r]: 0<=e<N_e and 0<=i,j<35 and 0<=x,r<3}",
        """
        y[i] = sum([r,j], J[x, r, e]*D[r,i,j]*u[e,j])
        """,
        [lp.GlobalArg("J,D,u", dtype=np.float64, shape=lp.auto), ...],
    )
    knl = t_unit.default_entrypoint

    knl = lp.split_reduction_inward(knl, "j")
    knl = lp_utils.hoist_invariant_multiplicative_terms_in_sum_reduction(
        knl, reduction_inames="j"
    )
    knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
        knl,
        within=None,
        subst_name="grad_without_jacobi_subst",
        arguments=variables("r i e"),
        terms_filter=lambda x: isinstance(x, Reduction),
    )

    transformed_t_unit = t_unit.with_kernel(knl)
    transformed_t_unit = lp.precompute(
        transformed_t_unit,
        "grad_without_jacobi_subst",
        sweep_inames=["r", "i"],
        precompute_outer_inames=frozenset({"e"}),
        temporary_address_space=lp.AddressSpace.PRIVATE,
    )

    x1 = lp.get_op_map(t_unit, subgroup_size=1).eval_and_sum({"N_e": 1})
    x2 = lp.get_op_map(transformed_t_unit, subgroup_size=1).eval_and_sum({"N_e": 1})

    assert x1 == 33075
    assert x2 == 7980  # i.e. this demonstrates a 4.14x reduction in flops



"""
.. currentmodule:: feinsum.codegen.loopy

.. autofunction:: generate_loopy
.. autofunction:: generate_loopy_with_opt_einsum_schedule
"""

from collections.abc import Mapping
from typing import Any, cast

import islpy as isl
import loopy as lp
import numpy as np
import pymbolic.primitives as p
from pytools import UniqueNameGenerator, memoize_on_first_arg

from feinsum.einsum import (
    INT_CLASSES,
    Argument,
    BatchedEinsum,
    ContractionSchedule,
    EinsumOperand,
    IntegralT,
    IntermediateResult,
    ShapeComponentT,
    ShapeT,
    SizeParam,
    get_opt_einsum_contraction_schedule,
    get_trivial_contraction_schedule,
)

LOOPY_LANG_VERSION = (2018, 2)


def _get_isl_basic_set(
    index_to_dim_length: Mapping[str, ShapeComponentT],
) -> isl.BasicSet:
    dim_name_to_ubound: dict[str, str | IntegralT] = {
        idx: dim.name if isinstance(dim, SizeParam) else dim
        for idx, dim in index_to_dim_length.items()
    }

    space = isl.Space.create_from_names(
        isl.DEFAULT_CONTEXT,
        set=sorted(dim_name_to_ubound),
        params=sorted(
            bound for bound in dim_name_to_ubound.values() if isinstance(bound, str)
        ),
    )
    bset = isl.BasicSet.universe(space)

    for dim_name, ubound in sorted(dim_name_to_ubound.items()):
        if isinstance(ubound, str):
            bset = bset.add_constraint(
                isl.Constraint.ineq_from_names(
                    space, {1: -1, ubound: 1, dim_name: -1}
                )
            )
        else:
            assert isinstance(ubound, INT_CLASSES)
            bset = bset.add_constraint(
                isl.Constraint.ineq_from_names(space, {1: ubound - 1, dim_name: -1})
            )

        bset = bset.add_constraint(
            isl.Constraint.ineq_from_names(space, {1: 0, dim_name: 1})
        )

    return bset


def make_subscript(
    name: str,
    idx_set: tuple[str, ...],
) -> p.Subscript:
    return p.Subscript(
        p.Variable(name),
        tuple(p.Variable(idx) for idx in idx_set),
    )


def make_access_expr(
    name: str,
    idx_set: tuple[str, ...],
) -> p.Call:
    return p.Call(
        p.Variable(_get_input_subst_name(name)),
        tuple(p.Variable(idx) for idx in idx_set),
    )


def _get_input_subst_name(x: str) -> str:
    return f"_fe_subst_{x}"


def _get_out_in_idx_sets(
    subscripts: str,
) -> tuple[tuple[str, ...], tuple[tuple[str, ...], ...]]:
    from feinsum.make_einsum import _normalize_einsum_subscript

    in_specs, out_spec = subscripts.split("->")
    out_idx_set = _normalize_einsum_subscript(out_spec, is_output=True)
    in_idx_sets = tuple(
        _normalize_einsum_subscript(in_spec, is_output=False)
        for in_spec in in_specs.split(",")
    )
    return out_idx_set, in_idx_sets


@memoize_on_first_arg
def generate_loopy(
    einsum: BatchedEinsum, schedule: ContractionSchedule | None = None
) -> "lp.TranslationUnit":
    """
    Returns a :class:`loopy.TranslationUnit` with the reductions scheduled by
    *contract_path*.

    :param schedule: An optional instance of
        :class:`~feinsum.einsum.ContractionSchedule`. Defaults to the trivial
        contraction schedule if not provided.
    """
    from functools import reduce

    if schedule is None:
        schedule = get_trivial_contraction_schedule(einsum)

    assert isinstance(schedule, ContractionSchedule)

    # {{{ prepare unique name generator

    vng = UniqueNameGenerator()
    vng.add_names(einsum.all_args)
    vng.add_names(einsum.all_indices)
    vng.add_names(p.name for p in einsum.all_size_params)

    # }}}

    statements: list[lp.Assignment] = []
    knl_args: list[lp.ArrayArg | lp.TemporaryVariable] = []
    substitutions: dict[str, lp.SubstitutionRule] = {}

    # {{{ populate operands in knl_args

    for arg_name in sorted(einsum.all_args):
        dtype = einsum.arg_to_dtype[arg_name]
        shape = tuple(
            dim.name if isinstance(dim, SizeParam) else dim
            for dim in einsum.arg_to_shape[arg_name]
        )
        knl_args.append(lp.GlobalArg(arg_name, dtype=dtype, shape=shape))

    # }}}

    # {{{ populate substitutions with arguments

    for arg in sorted(einsum.all_args):
        subst_name = _get_input_subst_name(arg)
        ndim = len(einsum.arg_to_shape[arg])
        subst = lp.SubstitutionRule(
            subst_name,
            tuple(f"d_{i}" for i in range(ndim)),
            make_subscript(arg, tuple(f"d_{i}" for i in range(ndim))),
        )
        substitutions[subst_name] = subst

    # }}}

    # {{{ start maintaining the inames involved.

    istepxindex_to_iname: dict[tuple[int, str], str] = {}
    iname_to_ubound: dict[str, ShapeComponentT] = {}

    for i_step, subscripts in enumerate(schedule.subscripts):
        _, in_idx_sets = _get_out_in_idx_sets(subscripts)
        all_indices = reduce(
            frozenset.union,
            (frozenset(idx_set) for idx_set in in_idx_sets),
            cast("frozenset[str]", frozenset()),
        )
        for idx in sorted(all_indices):
            istepxindex_to_iname[i_step, idx] = vng("i")

    sched_arg_to_shape: dict[Argument, ShapeT] = {}
    for ioperand, einsum_arg in enumerate(einsum.args[0]):
        sched_arg_to_shape[EinsumOperand(ioperand)] = einsum_arg.shape

    for i_step, (result_name, subscripts, operands) in enumerate(
        zip(
            schedule.result_names,
            schedule.subscripts,
            schedule.arguments,
            strict=True,
        )
    ):
        idx_to_length: dict[str, ShapeComponentT] = {}
        out_idx_set, in_idx_sets = _get_out_in_idx_sets(subscripts)

        for idx_set, operand in zip(in_idx_sets, operands, strict=True):
            for idx, axis_len in zip(
                idx_set, sched_arg_to_shape[operand], strict=True
            ):
                assert idx_to_length.setdefault(idx, axis_len) == axis_len

        iname_to_ubound.update(
            {
                istepxindex_to_iname[i_step, idx]: axis_len
                for idx, axis_len in idx_to_length.items()
            }
        )
        sched_arg_to_shape[IntermediateResult(result_name)] = tuple(
            idx_to_length[idx] for idx in out_idx_set
        )
    del sched_arg_to_shape

    # }}}

    # {{{ generate domains.

    domains: list[isl.BasicSet] = []

    for i_step in range(schedule.nsteps):
        inames = tuple(
            sorted(
                {
                    iname
                    for (istepxindex, iname) in istepxindex_to_iname.items()
                    if i_step == istepxindex[0]
                }
            )
        )
        domains.append(
            _get_isl_basic_set({iname: iname_to_ubound[iname] for iname in inames})
        )

    # }}}

    # {{{ generate statements and substitutions

    for args in einsum.args:
        dtypes = {arg.name: arg.dtype for arg in args}
        sched_arg_to_lp_name: dict[Argument, str] = {}

        for j, einsum_arg in enumerate(args):
            sched_arg_to_lp_name[EinsumOperand(j)] = einsum_arg.name

        for i_step, (result_name, operands, subscripts) in enumerate(
            zip(
                schedule.result_names,
                schedule.arguments,
                schedule.subscripts,
                strict=True,
            )
        ):
            lp_name = vng(result_name)
            lp_dtype = np.result_type(
                *[dtypes[sched_arg_to_lp_name[operand]] for operand in operands]
            )
            sched_arg_to_lp_name[IntermediateResult(result_name)] = lp_name
            dtypes[lp_name] = lp_dtype
            if result_name != schedule.result_names[-1]:
                knl_args.append(
                    lp.TemporaryVariable(
                        lp_name,
                        dtype=lp_dtype,
                        shape=lp.auto,
                        address_space=lp.AddressSpace.GLOBAL,
                    )
                )
            else:
                knl_args.append(lp.GlobalArg(lp_name, dtype=lp_dtype, shape=lp.auto))

            out_idx_set, in_idx_sets = _get_out_in_idx_sets(subscripts)
            out_idx_set = tuple(
                istepxindex_to_iname[i_step, idx] for idx in out_idx_set
            )
            in_idx_sets = tuple(
                tuple(istepxindex_to_iname[i_step, idx] for idx in idx_set)
                for idx_set in in_idx_sets
            )
            inames_to_sum_over = reduce(
                frozenset.union,
                (frozenset(idx_set) for idx_set in in_idx_sets),
                cast("frozenset[str]", frozenset()),
            ) - frozenset(out_idx_set)

            lhs = make_subscript(lp_name, out_idx_set)
            rhs: p.ExpressionNode = p.Product(
                tuple(
                    make_access_expr(sched_arg_to_lp_name[operand], in_idx_set)
                    for operand, in_idx_set in zip(
                        operands, in_idx_sets, strict=True
                    )
                )
            )
            if inames_to_sum_over:
                rhs = lp.Reduction(
                    "sum",
                    tuple(p.Variable(iname) for iname in inames_to_sum_over),
                    rhs,
                )

            statements.append(lp.Assignment(lhs, rhs))

            subst_name = _get_input_subst_name(lp_name)
            subst = lp.SubstitutionRule(
                subst_name,
                tuple(f"d_{i}" for i in range(len(out_idx_set))),
                make_subscript(
                    lp_name, tuple(f"d_{i}" for i in range(len(out_idx_set)))
                ),
            )
            substitutions[subst_name] = subst

    # }}}

    t_unit = lp.make_kernel(
        domains,
        statements,
        kernel_data=[*knl_args, ...],
        substitutions=substitutions,
        lang_version=LOOPY_LANG_VERSION,
    )
    return t_unit


def generate_loopy_with_opt_einsum_schedule(
    expr: BatchedEinsum, **opt_einsum_kwargs: Any
) -> "lp.TranslationUnit":
    """
    Returns a :class:`loopy.TranslationUnit` with the
    :class:`~feinsum.einsum.ContractionSchedule` specified via
    :func:`opt_einsum.contract_path`.

    :param opt_einsum_kwargs: kwargs to be passed to
        :func:`~feinsum.einsum.get_opt_einsum_contraction_schedule`.
    """
    return generate_loopy(expr, get_opt_einsum_contraction_schedule(expr))



from typing import Any

import loopy as lp
import numpy as np
from pyopencl.tools import (  # noqa
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)

import feinsum as f
from feinsum.tuning import IntParameter, transform_param


@transform_param("wg_size", lambda ensm: (IntParameter(8, 16), IntParameter(8, 16)))
def transform(
    t_unit: lp.TranslationUnit,
    wg_size: tuple[int, int],
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:
    ref_einsum = f.einsum("ijk->ij", f.array("A", ("I", 72, 4), np.float64))
    subst_map = f.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    i = subst_map["i"]
    j = subst_map["j"]

    l_0_size, l_1_size = wg_size
    assert isinstance(wg_size, tuple)
    assert 8 <= l_0_size <= 16 and 8 <= l_1_size <= 16

    t_unit = lp.split_iname(t_unit, i, l_1_size, inner_tag="l.1", outer_tag="g.0")
    t_unit = lp.split_iname(t_unit, j, l_0_size, inner_tag="l.0")

    return t_unit


def test_transform(ctx_factory):
    import os

    cl_ctx = ctx_factory()

    expr = f.einsum("ijk->ij", f.array("P", ("I", 72, 4), np.float64))
    f.autotune(expr, os.path.abspath(__file__), cl_ctx, stop_after=3,
               long_dim_length=100)



from __future__ import annotations

__doc__ = """
.. autofunction:: query
.. autofunction:: get_timed_einsums_in_db
.. autofunction:: record_into_db

.. autoclass:: QueryInfo
"""

import json
import logging
import os
import sqlite3
from dataclasses import dataclass
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
)

import numpy as np
import numpy.typing as npt
from immutables import Map

from feinsum.einsum import (
    INT_CLASSES,
    BatchedEinsum,
    ShapeComponentT,
    SizeParam,
)

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    # avoid making pyopencl a hard dep.
    from collections.abc import Callable, Mapping, Sequence

    import loopy as lp
    import pyopencl as cl

    from feinsum.cl_utils import ContextT, DeviceT

    # transform: (t_unit, insn_match, kernel_name)
    TransformT = Callable[
        [lp.TranslationUnit, Any | None, str | None], lp.TranslationUnit
    ]


DEFAULT_DB = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "data",
    "transform_archive_v6.sqlite",
)
TIMINGS_TABLENAME = "FEINSUM_TIMING_FACTS"


def dump_arg_to_dtype(einsum: BatchedEinsum) -> str:
    return json.dumps(
        {arg: dtype.name for arg, dtype in einsum.arg_to_dtype.items()},
        sort_keys=True,
    )


def dump_index_to_length(einsum: BatchedEinsum) -> str:
    return json.dumps(
        {
            k: v
            for k, v in einsum.index_to_dim_length.items()
            if isinstance(v, INT_CLASSES)
        },
        sort_keys=True,
    )


def dump_arg_names(einsum: BatchedEinsum) -> str:
    arg_names = [[arg.name for arg in arg_row] for arg_row in einsum.args]
    return json.dumps(arg_names)


def dump_cl_version(cl_device: cl.Device) -> str:
    # TODO: needs to consider more things into account
    return f"{cl_device.vendor}-{cl_device.driver_version}"


def dump_op_info(einsum: BatchedEinsum, long_dim_length: int) -> str:
    from pymbolic.mapper.evaluator import evaluate_to_float

    from feinsum.measure import _get_giga_ops_from_einsum

    eval_context = {
        dim.name: long_dim_length
        for dim in einsum.index_to_dim_length.values()
        if isinstance(dim, SizeParam)
    }
    dtype_to_ops = {
        k: evaluate_to_float(v, eval_context)
        for k, v in _get_giga_ops_from_einsum(einsum).items()
    }
    return json.dumps({k.name: v for k, v in dtype_to_ops.items()}, sort_keys=True)


def load_op_info(op_info: str) -> Map[np.dtype[Any], float]:
    return Map({np.dtype(k): v for k, v in json.loads(op_info).items()})


def _process_param(param: Any) -> Any:
    if isinstance(param, (int, bool)):
        return param
    elif isinstance(param, list):
        return tuple(_process_param(k) for k in param)
    else:
        raise NotImplementedError(type(param))


def load_transform_params(params_str: str) -> Map[str, Any]:
    preprocessed_params = json.loads(params_str)
    assert isinstance(preprocessed_params, dict)
    assert all(isinstance(k, str) for k in preprocessed_params)

    return Map({k: _process_param(v) for k, v in preprocessed_params.items()})


def dump_device_name(cl_device: DeviceT) -> str:
    dev_name = cl_device.name
    assert isinstance(dev_name, str)
    return (
        dev_name.replace(" ", "_")
        .replace("-", "_")
        .replace("@", "AT")
        .replace("(", "_")
        .replace(")", "_")
        .replace(".", "DOT")
    )


@dataclass(frozen=True)
class QueryInfo:
    transform_id: str
    transform_params: Map[str, Any]
    runtime_in_sec: float
    compiler_version: str
    giga_op_info: Map[np.dtype[Any], float]
    _einsum: BatchedEinsum

    def giga_op_rate(self, dtype: npt.DTypeLike) -> float:
        return self.giga_op_info[np.dtype(dtype)] / self.runtime_in_sec

    @cached_property
    def transform(self) -> TransformT:
        from feinsum.tuning import (
            _get_impls_path,
            get_transform_func_from_module_path,
        )

        module_path = os.path.join(_get_impls_path(), self.transform_id)
        return get_transform_func_from_module_path(module_path).bind_args(
            self._einsum, **self.transform_params
        )


def query(
    einsum: BatchedEinsum,
    cl_ctx: ContextT,
    *,
    database: str = DEFAULT_DB,
    err_if_no_results: bool = False,
) -> tuple[QueryInfo, ...]:
    """
    Returns facts of previous recorded runs of *einsum* on *cl_ctx*.

    :param err_if_no_results: If *True*, raises a :class:`RuntimeError` if no
        recorded runs corresponding to *einsum* are available in *database*.
        Defaults to *False*.
    """
    from feinsum.canonicalization import canonicalize_einsum

    einsum = canonicalize_einsum(einsum)
    conn = sqlite3.connect(database)
    cursor = conn.cursor()

    if len({dev.name for dev in cl_ctx.devices}) > 1:
        raise NotImplementedError(
            "CL contexts with multiple types of devices" " not supported."
        )

    cl_device = cl_ctx.devices[0]
    device_name = dump_device_name(cl_device)
    subscripts = einsum.get_subscripts()
    index_to_length = dump_index_to_length(einsum)
    arg_names = dump_arg_names(einsum)
    arg_to_dtype = dump_arg_to_dtype(einsum)

    cursor.execute(
        " SELECT name FROM sqlite_master" " WHERE (type='table' AND name=?);",
        (TIMINGS_TABLENAME,),
    )

    if not cursor.fetchall():
        raise RuntimeError(
            f"Database '{database}' does not" " contain the timing facts table."
        )

    cursor.execute(
        " SELECT"
        "     transform_id,"
        "     transform_params,"
        "     runtime_in_sec,"
        "     compiler_version,"
        "     giga_op_info"
        "  FROM "
        f"    {TIMINGS_TABLENAME}"
        " WHERE ("
        "    subscripts = ?"
        "    AND index_to_length = ?"
        "    AND arg_names = ?"
        "    AND arg_to_dtype = ?"
        "    AND device_name = ?"
        ");",
        (subscripts, index_to_length, arg_names, arg_to_dtype, device_name),
    )

    facts = cursor.fetchall()

    query_result = tuple(
        QueryInfo(
            transform_id=fact[0],
            transform_params=json.loads(fact[1]),
            runtime_in_sec=fact[2],
            compiler_version=fact[3],
            giga_op_info=load_op_info(fact[4]),
            _einsum=einsum,
        )
        for fact in facts
    )
    conn.close()

    if not query_result and err_if_no_results:
        str_idx_to_size = ", ".join(
            f"{idx}: {lngth}"
            for idx, lngth in (einsum.index_to_dim_length.items())
            if not isinstance(lngth, SizeParam)
        )
        stringified_einsum = (
            f"{einsum.get_subscripts()} [{str_idx_to_size}]"
            f" [#outputs={einsum.b}]"
        )
        raise RuntimeError(
            "No facts found for the einsum:" f" `{stringified_einsum}`."
        )

    return query_result


def _get_batched_einsum_from_sql_row(
    subscripts: str,
    index_to_length: Mapping[str, ShapeComponentT],
    arg_names: Sequence[Sequence[str]],
    arg_to_dtype: Mapping[str, str],
) -> BatchedEinsum:
    from functools import reduce
    from typing import cast

    from feinsum.einsum import SizeParam
    from feinsum.make_einsum import (
        _normalize_einsum_subscript,
        array,
        batched_einsum,
    )

    in_specs, _ = subscripts.split("->")
    index_to_length = dict(index_to_length)
    in_idx_sets = tuple(
        _normalize_einsum_subscript(in_spec, is_output=False)
        for in_spec in in_specs.split(",")
    )

    all_indices = reduce(
        frozenset.union,
        (frozenset(in_idx_set) for in_idx_set in in_idx_sets),
        cast("frozenset[str]", frozenset()),
    )
    for idx in all_indices:
        if idx not in index_to_length:
            assert idx.islower()
            index_to_length[idx] = SizeParam(idx.upper())

    arg_to_shape = {
        arg: [index_to_length[idx] for idx in in_idx_set]
        for arg_row in arg_names
        for in_idx_set, arg in zip(in_idx_sets, arg_row, strict=True)
    }
    args = [
        [array(arg, arg_to_shape[arg], arg_to_dtype[arg]) for arg in arg_row]
        for arg_row in arg_names
    ]
    return batched_einsum(subscripts, args)


def get_timed_einsums_in_db(
    cl_device: DeviceT, database: str = DEFAULT_DB
) -> tuple[BatchedEinsum, ...]:
    r"""
    Returns a :class:`tuple` of :class:`~feinsum.einsum.BatchedEinsum`\ s for
    which some timing data is available on the OpenCL device *device* in the
    database *database*.
    """

    device_name = dump_device_name(cl_device)

    conn = sqlite3.connect(database)
    cursor = conn.cursor()
    cursor.execute(
        " SELECT"
        "     subscripts,"
        "     index_to_length,"
        "     args,"
        "     arg_to_dtype"
        "  FROM "
        f"    {TIMINGS_TABLENAME}"
        " WHERE "
        "    device_name = ?"
        ";",
        (device_name,),
    )

    facts = set(cursor.fetchall())
    seen_einsums: list[BatchedEinsum] = []
    conn.close()

    for subscripts, index_to_length_str, arg_names_str, arg_to_dtype_str in facts:
        index_to_length = json.loads(index_to_length_str)
        arg_to_dtype = json.loads(arg_to_dtype_str)
        arg_names = json.loads(arg_names_str)
        seen_einsums.append(
            _get_batched_einsum_from_sql_row(
                subscripts, index_to_length, arg_names, arg_to_dtype
            )
        )

    # Asserts that the canonicalization was sound.
    assert len(set(seen_einsums)) == len(seen_einsums)

    return tuple(seen_einsums)


def _create_timings_table_if_non_existent(conn: sqlite3.Connection) -> None:
    cursor = conn.cursor()
    cursor.execute(
        " SELECT name FROM sqlite_master" " WHERE (type='table' AND name=?);",
        (TIMINGS_TABLENAME,),
    )

    if not cursor.fetchall():
        # device table not available
        logger.info(f"Table {TIMINGS_TABLENAME} not in DB, creating one.")
        cursor.execute(
            f"CREATE TABLE {TIMINGS_TABLENAME} ("
            " ID INTEGER PRIMARY KEY AUTOINCREMENT,"
            " subscripts TEXT,"
            " index_to_length TEXT,"
            " args TEXT,"
            " arg_to_dtype TEXT,"
            " device_name TEXT,"
            " transform_id TEXT,"
            " transform_params TEXT,"
            " runtime_in_sec REAL,"
            " compiler_version TEXT,"
            " giga_op_info TEXT,"
            " timestamp TEXT"
            ")"
        )
    conn.commit()


def record_into_db(
    einsum: BatchedEinsum,
    cl_ctx: cl.Context,
    module_path: str,
    transform_params: Mapping[str, Any],
    database: str | sqlite3.Connection = DEFAULT_DB,
    long_dim_length: int = 100_000,
) -> None:
    """
    Records facts corresponding to the execution of *einsum* on *cl_ctx* with
    the transformation in *module_path* along with *transform_params* and
    records it in the SQL database *database*.
    """
    from feinsum.canonicalization import canonicalize_einsum
    from feinsum.measure import stringify_comparison_vs_roofline, timeit
    from feinsum.tuning import _get_impls_path, get_transform_func_from_module_path

    dirpath, transform_space_id = os.path.split(module_path)
    if dirpath != _get_impls_path():
        transform_space_id = module_path
    einsum = canonicalize_einsum(einsum)

    transform_func = get_transform_func_from_module_path(module_path).bind_args(
        einsum, **transform_params
    )
    logger.info(
        "\n"
        + stringify_comparison_vs_roofline(
            einsum,
            transform=transform_func,
            cl_ctx=cl_ctx,
        )
    )
    runtime = timeit(
        einsum,
        cl_ctx=cl_ctx,
        transform=transform_func,
        long_dim_length=long_dim_length,
    )
    if isinstance(database, str):
        conn = sqlite3.connect(database)
    else:
        assert isinstance(database, sqlite3.Connection)
        conn = database

    _create_timings_table_if_non_existent(conn)
    cursor = conn.cursor()
    subscripts = einsum.get_subscripts()
    index_to_length = dump_index_to_length(einsum)
    arg_names = dump_arg_names(einsum)
    arg_to_dtype = dump_arg_to_dtype(einsum)
    transform_params_str = json.dumps(transform_params, sort_keys=True)
    (cl_device,) = cl_ctx.devices
    device_name = dump_device_name(cl_device)
    compiler_version = dump_cl_version(cl_device)
    op_info = dump_op_info(einsum, long_dim_length=long_dim_length)

    # {{{ compute timestamp in Chicago

    from datetime import datetime

    import pytz

    timestamp = datetime.now(pytz.timezone("America/Chicago")).strftime(
        "%Y_%m_%d_%H%M%S"
    )

    # }}}

    cursor.execute(
        f"INSERT INTO {TIMINGS_TABLENAME}"
        " (subscripts, index_to_length, args,"
        "  arg_to_dtype, device_name, transform_id,"
        "  transform_params, runtime_in_sec,"
        "  compiler_version, giga_op_info, timestamp)"
        " VALUES (?,?,?,?,?,?,?,?,?,?,?)",
        (
            subscripts,
            index_to_length,
            arg_names,
            arg_to_dtype,
            device_name,
            transform_space_id,
            transform_params_str,
            runtime,
            compiler_version,
            op_info,
            timestamp,
        ),
    )

    conn.commit()


# vim: foldmethod=marker



import logging

import loopy as lp
import numpy as np
import pyopencl as cl

import feinsum as f

logging.basicConfig(level="INFO")
logger = logging.getLogger(__name__)


def get_grad_einsum(ndofs, ndim):
    return f.einsum(
        "xre,rij,ej->xei",
        f.array((ndim, ndim, np.inf), "float64"),
        f.array((ndim, ndofs, ndofs), "float64"),
        f.array((np.inf, ndofs), "float64"),
        arg_names=["J", "D", "u"],
    )


def variant_0(t_unit):
    # The one in `main` right now
    ref_einsum = get_grad_einsum(ndofs=35, ndim=3)
    subst_map = f.match_t_unit_to_einsum(t_unit, ref_einsum)
    e, i = subst_map["e"], subst_map["i"]
    return lp.split_iname(
        lp.split_iname(t_unit, e, 8, outer_tag="g.0", inner_tag="l.1"),
        i,
        4,
        inner_tag="l.0",
    )


def variant_1(t_unit, nwork_items_per_cell=4, ncells_per_workgroup=8):
    ref_einsum = get_grad_einsum(ndofs=35, ndim=3)
    subst_map = f.match_t_unit_to_einsum(t_unit, ref_einsum)

    r = subst_map["r"]
    i = subst_map["i"]
    j = subst_map["j"]
    e = subst_map["e"]
    J = subst_map["J"]
    x = subst_map["x"]

    t_unit = lp.split_reduction_outward(t_unit, r)
    t_unit = lp.realize_reduction(t_unit)

    t_unit = lp.split_iname(t_unit, i, nwork_items_per_cell)
    t_unit = lp.split_iname(t_unit, e, ncells_per_workgroup)

    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, r, only_var_names={f"acc_{j}"}
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        (r,),
        within=f"id:insn_{r}_update_*",
        new_inames=["r_evaluate_at_point"],
    )

    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, i + "_outer", only_var_names={f"acc_{r}"}
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        inames=(i + "_outer",),
        within="id:insn_r_init",
        new_inames=["i_outer_init"],
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        inames=(i + "_outer",),
        within="id:insn",
        new_inames=["i_outer_store"],
    )

    t_unit = lp.add_prefetch(
        t_unit,
        J,
        sweep_inames={"r_evaluate_at_point"},
        fetch_outer_inames=frozenset({x, i + "_inner", r + "_outer", e + "_inner"}),
        temporary_address_space=lp.AddressSpace.PRIVATE,
        temporary_name="J_prftch",
    )

    t_unit = lp.tag_inames(
        t_unit, {e + "_outer": "g.0", e + "_inner": "l.1", i + "_inner": "l.0"}
    )
    return t_unit


def variant_2(t_unit, nwork_items_per_cell=4, ncells_per_workgroup=8):
    t_unit = variant_1(
        t_unit,
        nwork_items_per_cell=nwork_items_per_cell,
        ncells_per_workgroup=ncells_per_workgroup,
    )
    t_unit = lp.add_prefetch(
        t_unit,
        "R",
        sweep_inames={"r_evaluate_at_point", "i_inner", "i_outer", "j"},
        fetch_outer_inames=frozenset({"e_outer"}),
        temporary_address_space=lp.AddressSpace.LOCAL,
        default_tag=None,
        temporary_name="R_prftch",
    )

    t_unit = lp.join_inames(t_unit, ["R_dim_0", "R_dim_1", "R_dim_2"], "iRprftch")
    t_unit = lp.split_iname(
        t_unit, "iRprftch", nwork_items_per_cell * ncells_per_workgroup
    )
    t_unit = lp.split_iname(
        t_unit,
        "iRprftch_inner",
        nwork_items_per_cell,
        inner_tag="l.0",
        outer_tag="l.1",
    )
    return t_unit


def variant_3(t_unit):
    from loopy.symbolic import Reduction

    ncells_per_workgroup = 9
    nworkitems_per_cell = 7
    j_tile_len = 9
    i_tile_len = 35

    ref_einsum = get_grad_einsum(ndofs=35, ndim=3)

    subst_map = f.match_t_unit_to_einsum(t_unit, ref_einsum)

    r = subst_map["r"]
    x = subst_map["x"]
    j = subst_map["j"]
    i = subst_map["i"]
    e = subst_map["e"]
    J = subst_map["J"]
    u = subst_map["u"]
    D = subst_map["D"]

    # {{{ term hoisting to match the flop count of opt_einsum

    knl = t_unit.default_entrypoint

    knl = lp.split_reduction_inward(knl, j)
    knl = lp.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, j)
    knl = lp.extract_multiplicative_terms_in_sum_reduction_as_subst(
        knl,
        within=None,
        subst_name="subst",
        arguments=[r, e, i],
        terms_filter=lambda x_: isinstance(x_, Reduction),
    )

    t_unit = t_unit.with_kernel(knl)

    # }}}

    t_unit = lp.split_iname(t_unit, i, i_tile_len, outer_iname="i_tile")
    t_unit = lp.split_iname(t_unit, j, j_tile_len, outer_iname="j_tile")

    t_unit = lp.rename_iname(t_unit, "i_inner", i)
    t_unit = lp.rename_iname(t_unit, "j_inner", j)
    t_unit = lp.split_iname(
        t_unit,
        e,
        ncells_per_workgroup,
        inner_iname="e_inner",
        outer_iname="e_outer",
        outer_tag="g.0",
        inner_tag="l.1",
    )
    t_unit = lp.split_iname(
        t_unit,
        "i",
        nworkitems_per_cell,
        inner_iname="i_inner",
        outer_iname="i_outer",
        inner_tag="l.0",
    )

    t_unit = lp.precompute(
        t_unit,
        J,
        sweep_inames=[r, x],
        precompute_outer_inames=frozenset(["e_inner", "e_outer", "i_inner"]),
        temporary_address_space=lp.AddressSpace.PRIVATE,
        temporary_name="J_prftch",
    )

    # {{{ TODO: Make precompute smarter (should be a single precompute call)

    t_unit = lp.precompute(
        t_unit,
        "subst",
        sweep_inames=["r"],
        precompute_outer_inames=frozenset(
            {
                "e_inner",
                "e_outer",
                "i_inner",
                "i_outer",
                "i_tile",
            }
        ),
        temporary_name="tmp_hoist",
        temporary_address_space=lp.AddressSpace.PRIVATE,
        compute_insn_id="insn_hoist",
    )
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, "i_outer", only_var_names=["tmp_hoist"]
    )

    t_unit = lp.duplicate_inames(t_unit, "i_outer", "id:insn_hoist", "i_outer_hoist")

    # }}}

    # {{{ Move 'u ' to shared.

    # Prefetch 'u' within the tile
    t_unit = lp.precompute(
        t_unit,
        u,
        sweep_inames=["e_inner", j],
        precompute_outer_inames=frozenset(["e_outer", "i_tile", "j_tile"]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        precompute_inames=["e_prftch", "j_prftch"],
        default_tag=None,
    )

    t_unit = lp.split_iname(
        t_unit, "e_prftch", ncells_per_workgroup, inner_tag="l.1", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, "j_prftch", nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )

    # }}}

    # {{{ Move 'R' to shared.

    t_unit = lp.add_prefetch(
        t_unit,
        D,
        sweep_inames=["r_0", "i_inner", "i_outer_hoist", "j"],
        fetch_outer_inames=frozenset(["e_outer", "i_tile", "j_tile"]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        dim_arg_names=["r_prftch", "i_prftch", "j_prftch"],
        default_tag=None,
    )
    if 0:
        # This branch improves perf. by 20% but, something in loopy
        # non-deterministically leads to very ugly domains.
        t_unit = lp.join_inames(
            t_unit, ["r_prftch", "i_prftch", "j_prftch"], "i_Rprftch"
        )

        t_unit = lp.split_iname(
            t_unit,
            "i_Rprftch",
            ncells_per_workgroup * nworkitems_per_cell,
            outer_tag="unr",
        )

        t_unit = lp.split_iname(
            t_unit,
            "i_Rprftch_inner",
            nworkitems_per_cell,
            inner_tag="l.0",
            outer_tag="l.1",
        )
    else:
        t_unit = lp.tag_inames(t_unit, "r_prftch:unr")
        t_unit = lp.split_iname(
            t_unit,
            "i_prftch",
            ncells_per_workgroup,
            inner_tag="l.1",
            outer_tag="unr",
        )
        t_unit = lp.split_iname(
            t_unit, "j_prftch", nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
        )

    # }}}

    logger.info("Done with prefetching 'R'.")

    # {{{ make buffer array smarter (should be a single call to buffer_array)

    t_unit = lp.buffer_array(
        t_unit,
        "_fe_out",
        buffer_inames=["x"],
        init_expression="0",
        default_tag=None,
        temporary_scope=lp.AddressSpace.PRIVATE,
    )
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, "i_outer", only_var_names={"_fe_out_buf"}
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames=["i_outer"],
        within="id:init__fe_out",
        new_inames=["_fe_out_init_1"],
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames=["i_outer"],
        within="id:store__fe_out",
        new_inames=["_fe_out_store_1"],
    )

    # }}}

    # {{{ must be smarter way of doing this in loopy

    t_unit = lp.realize_reduction(t_unit, insn_id_filter="insn_hoist")
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, frozenset(["r_0", "i_outer_hoist"]), only_var_names={"acc_j_tile_j"}
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["i_outer_hoist", "r_0"],
        within="id:insn_hoist_j_tile_j_init",
        new_inames=["i_outer_hoist_init", "r_0_init"],
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        ["i_outer_hoist", "r_0"],
        within="id:insn_hoist",
        new_inames=["i_outer_hoist_store", "r_0_store"],
    )

    # }}}

    logger.info("Done with transformations.")

    return t_unit


def paranumal_transform(t_unit, insn_match=None, kernel_name=None):
    from loopy.symbolic import Reduction
    from pymbolic import variables

    ref_einsum = get_grad_einsum(35, 3)
    kernel_name = kernel_name or t_unit.default_entrypoint.name

    subst_map = f.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    j = subst_map["j"]
    r = subst_map["r"]
    e = subst_map["e"]
    i = subst_map["i"]
    u = subst_map["u"]
    x = subst_map["x"]
    J = subst_map["J"]

    knl = t_unit[kernel_name]
    knl = lp.split_reduction_inward(knl, j)
    knl = lp.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, j)
    knl = lp.extract_multiplicative_terms_in_sum_reduction_as_subst(
        knl,
        within=None,
        subst_name="subst",
        arguments=variables(f"{r} {e} {i}"),
        terms_filter=lambda x: isinstance(x, Reduction),
    )
    t_unit = t_unit.with_kernel(knl)

    NblockV = 4

    t_unit = lp.split_iname(t_unit, e, NblockV, inner_tag="l.1", outer_tag="g.0")
    t_unit = lp.tag_inames(t_unit, {i: "l.0"})
    t_unit = lp.precompute(
        t_unit,
        u,
        sweep_inames=[f"{e}_inner", j],
        precompute_inames=["eprftch_u", "jprftch_u"],
        temporary_address_space=lp.AddressSpace.LOCAL,
        default_tag=None,
    )
    t_unit = lp.tag_inames(t_unit, {"eprftch_u": "l.1", "jprftch_u": "l.0"})
    t_unit = lp.precompute(
        t_unit,
        "subst",
        sweep_inames=[r],
        precompute_outer_inames=frozenset({f"{e}_outer", f"{e}_inner", i}),
        temporary_address_space=lp.AddressSpace.PRIVATE,
        default_tag="unr",
    )

    t_unit = lp.precompute(
        t_unit,
        J,
        sweep_inames=[x, r],
        precompute_outer_inames=frozenset({f"{e}_outer", f"{e}_inner", i}),
        temporary_address_space=lp.AddressSpace.PRIVATE,
        default_tag="unr",
    )
    t_unit = lp.tag_inames(t_unit, {r: "unr", j: "unr", x: "unr"})

    return t_unit


def main():
    cl_ctx = cl.create_some_context()

    expr = get_grad_einsum(ndofs=35, ndim=3)
    print(
        f.stringify_comparison_vs_roofline(
            expr,
            cl_ctx=cl_ctx,
            transform=paranumal_transform,
            long_dim_length=1000,
            ignore_unknown_device=True,  # For CI
        )
    )


if __name__ == "__main__":
    main()



import logging
import math
from typing import Any

import loopy as lp
import numpy as np
from more_itertools import zip_equal as szip

import feinsum as fnsm
from feinsum.einsum import INT_CLASSES, SizeParam
from feinsum.tuning import IntParameter, einsum_arg, transform_param
from feinsum.utils import get_n_redn_dim

logger = logging.getLogger(__name__)

# Values for Volta micro-arch. Generalize it via a class
MAX_SHARED_MEM_PER_WG = 48e3  # in bytes
REG_FILE_SPACE_PER_WI = 256 * 4  # in bytes


def _is_ensm_tensor_contraction(ensm: fnsm.BatchedEinsum) -> bool:
    # TC requires noperands == 2
    if ensm.n != 2 or ensm.b != 1:
        return False

    in_idx_set1, in_idx_set2 = ensm.in_idx_sets

    # TC requires each out index be indexed in exactly one operand
    for out_idx in ensm.out_idx_set:
        if not ((out_idx in in_idx_set1) ^ (out_idx in in_idx_set2)):
            return False
    if (len(set(in_idx_set1)) != len(in_idx_set1)) or (
        len(set(in_idx_set2)) != len(in_idx_set2)
    ):
        return False

    all_redn_indices = (frozenset(in_idx_set1) | frozenset(in_idx_set2)) - frozenset(
        ensm.out_idx_set
    )
    # TC requires all reduction indices be present in both the operands
    for redn_idx in all_redn_indices:
        if redn_idx not in in_idx_set1 or redn_idx not in in_idx_set2:
            return False

    return True


def _get_indices(
    ensm: fnsm.BatchedEinsum,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    return (
        ensm.out_idx_set,
        tuple(
            idx
            for idx in ensm.in_idx_sets[0] + ensm.in_idx_sets[1]
            if idx not in ensm.out_idx_set
        ),
    )


def _get_operand_names(ensm: fnsm.BatchedEinsum) -> tuple[str, str]:
    assert ensm.b == 1 and ensm.n == 2
    ((arg1, arg2),) = ensm.args
    return (arg1.name, arg2.name)


@einsum_arg("ensm", lambda ensm: ensm)
@transform_param(
    "i_axis_mapping_perm",
    lambda ensm: IntParameter(0, math.perm(ensm.ndim, min(ensm.ndim, 4)) - 1),
)
@transform_param(
    "output_tile_lengths",
    lambda ensm: tuple(IntParameter(1, 32) for _ in range(min(ensm.ndim, 4))),
)
@transform_param(
    "t_redns",
    lambda ensm: tuple(IntParameter(1, 16) for i in range(get_n_redn_dim(ensm))),
)
def transform(
    t_unit: lp.TranslationUnit,
    ensm: fnsm.BatchedEinsum,
    i_axis_mapping_perm: int,
    output_tile_lengths: tuple[int, ...],
    t_redns: tuple[int, ...],
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:
    if ensm.ndim < 2:
        raise ValueError(
            "This algorithm needs al least two dimensions" " in the output array."
        )
    if len(ensm.all_size_params) != 0:
        raise NotImplementedError("Parametric lengths are not allowed.")

    import itertools

    i_tx: int
    i_ty: int
    i_rx: int | None
    i_ry: int | None

    for i, index_mapping in enumerate(
        itertools.permutations(list(range(ensm.ndim)), min(4, ensm.ndim))
    ):
        if i == i_axis_mapping_perm:
            if len(index_mapping) == 2:
                i_tx, i_ty = index_mapping
                i_rx = None
                i_ry = None
            elif len(index_mapping) == 3:
                i_tx, i_ty, i_rx = index_mapping
                i_ry = None
            elif len(index_mapping) == 4:
                i_tx, i_ty, i_rx, i_ry = index_mapping
            else:
                raise AssertionError()
            break
    else:
        raise AssertionError(
            f"{i_axis_mapping_perm} is an invalid" " permutation index."
        )
    tx: int
    ty: int
    rx: int | None
    ry: int | None
    if len(output_tile_lengths) == 4:
        tx, ty, rx, ry = output_tile_lengths
    elif len(output_tile_lengths) == 3:
        ry = None
        tx, ty, rx = output_tile_lengths
    elif len(output_tile_lengths) == 2:
        rx, ry = None, None
        tx, ty = output_tile_lengths
    else:
        raise ValueError(
            "len(output_tile_lengths) can be either 2, 3, or 4;"
            f" got {len(output_tile_lengths)}."
        )

    import loopy.match as lp_match

    kernel_name = kernel_name or t_unit.default_entrypoint.name
    knl = t_unit[kernel_name]
    within = lp_match.parse_match(insn_match)
    (insn_id,) = [insn.id for insn in knl.instructions if within(knl, insn)]
    del knl

    if not _is_ensm_tensor_contraction(ensm):
        raise ValueError(f"{ensm} is not a tensor contraction.")

    ensm_free_indices, ensm_redn_indices = _get_indices(ensm)
    ensm_A, ensm_B = _get_operand_names(ensm)

    # {{{ sanity checks on param space

    # Verify legality of tile lengths
    # -------------------------------
    i_tx_dim_length = ensm.index_to_dim_length[ensm.out_idx_set[i_tx]]
    if isinstance(i_tx_dim_length, INT_CLASSES) and tx > i_tx_dim_length:
        raise fnsm.InvalidParameterError("Tx > Nx")

    i_ty_dim_length = ensm.index_to_dim_length[ensm.out_idx_set[i_ty]]
    if isinstance(i_ty_dim_length, INT_CLASSES) and ty > i_ty_dim_length:
        raise fnsm.InvalidParameterError("Ty > Ny")

    if i_rx is not None:
        i_rx_dim_length = ensm.index_to_dim_length[ensm.out_idx_set[i_rx]]
        assert rx is not None
        if isinstance(i_rx_dim_length, INT_CLASSES) and rx > i_rx_dim_length:
            raise fnsm.InvalidParameterError("Rx > Nx")

    if i_ry is not None:
        i_ry_dim_length = ensm.index_to_dim_length[ensm.out_idx_set[i_ry]]
        assert ry is not None
        if isinstance(i_ry_dim_length, INT_CLASSES) and ry > i_ry_dim_length:
            raise fnsm.InvalidParameterError("Ry > Ny")

    for redn_idx, redn_tile in zip(ensm_redn_indices, t_redns, strict=True):
        redn_idx_dim_length = ensm.index_to_dim_length[redn_idx]
        if (
            isinstance(redn_idx_dim_length, INT_CLASSES)
            and redn_tile > redn_idx_dim_length
        ):
            raise fnsm.InvalidParameterError(
                f"redn_dim {redn_idx}'s tile is greater the axis length."
            )

    # Verify Tx, Ty, Rx, Ry are distinct dimensions
    # ---------------------------------------------
    if len({i_rx, i_tx, i_ry, i_ty, None}) != min(4, ensm.ndim) + 1:
        raise AssertionError("i_rx, i_tx, i_ty, i_ry are not distinct.")

    # Verify shared memory
    # --------------------
    from pytools import product

    l_sm_a, l_sm_b = [
        product(
            tile_length
            for tile_length, i_tile in [(tx, i_tx), (ty, i_ty)]
            if ensm.out_idx_set[i_tile] in idx_set
        )
        * product(t_redns)
        for idx_set in ensm.in_idx_sets
    ]
    if (
        l_sm_a * ensm.arg_to_dtype[ensm_A].itemsize
        + l_sm_b * ensm.arg_to_dtype[ensm_B].itemsize
    ) > MAX_SHARED_MEM_PER_WG:
        raise fnsm.InvalidParameterError("Exceeds local memory limits")

    # Verify register file usage
    # --------------------------
    if (rx or 1) * (ry or 1) * np.result_type(
        ensm.arg_to_dtype[ensm_A], ensm.arg_to_dtype[ensm_B]
    ).itemsize > REG_FILE_SPACE_PER_WI:
        raise fnsm.InvalidParameterError("Exceeds register file limits")

    assert tx <= 32 and ty <= 32

    # Note: we do not restrict the tile-lengths(tx, ty) as in this case the
    # tile-lengths also correspond to the execution grid size.
    new_t_redns = []
    for redn_idx, t_redn in zip(ensm_redn_indices, t_redns, strict=True):
        axis_len = ensm.index_to_dim_length[redn_idx]
        if isinstance(axis_len, SizeParam):
            new_t_redns.append(t_redn)
        else:
            assert isinstance(axis_len, INT_CLASSES)
            n_tiles = math.ceil(axis_len / t_redn)
            new_t_redns.append(math.ceil(axis_len / n_tiles))

    assert len(new_t_redns) == len(t_redns)
    assert all(
        new_t_redn <= old_t_redn
        for new_t_redn, old_t_redn in zip(new_t_redns, t_redns, strict=False)
    )
    t_redns = tuple(new_t_redns)
    del new_t_redns

    # }}}

    # type-ignore-reason: mypy is correct, but we have asserted earlier that we
    # won't accept parametric dim lengths
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit,
        ensm,
        insn_match=insn_match,
        kernel_name=kernel_name,
        long_dim_length=max(  # type: ignore[type-var,operator]
            ensm.index_to_dim_length.values(), default=1
        )
        + 1,
    )
    vng = t_unit[kernel_name].get_var_name_generator()
    ing = t_unit[kernel_name].get_instruction_id_generator()

    A = subst_map[ensm_A]
    B = subst_map[ensm_B]
    # C = subst_map["_fe_out"]
    free_indices = tuple(subst_map[idx] for idx in ensm_free_indices)
    redn_indices = tuple(subst_map[idx] for idx in ensm_redn_indices)
    tx_inner, tx_outer = vng("tx_inner"), vng("tx_outer")
    ty_inner, ty_outer = vng("ty_inner"), vng("ty_outer")
    rx_inner, rx_outer = vng("rx_inner"), vng("rx_outer")
    ry_inner, ry_outer = vng("ry_inner"), vng("ry_outer")
    inner_redn_inames = tuple(
        vng(f"{redn_iname}_inner") for redn_iname in redn_indices
    )
    outer_redn_inames = tuple(
        vng(f"{redn_iname}_outer") for redn_iname in redn_indices
    )

    iouter = vng("iouter")
    a_prftch_insn_id = ing("a_prftch_insn")
    b_prftch_insn_id = ing("b_prftch_insn")

    # {{{ split free indices

    t_unit = lp.split_iname(
        t_unit,
        free_indices[i_tx],
        tx,
        inner_iname=tx_inner,
        outer_iname=tx_outer,
        inner_tag="l.0",
    )
    t_unit = lp.split_iname(
        t_unit,
        free_indices[i_ty],
        ty,
        inner_iname=ty_inner,
        outer_iname=ty_outer,
        inner_tag="l.1",
    )
    if i_rx is not None:
        assert rx is not None
        t_unit = lp.split_iname(
            t_unit,
            free_indices[i_rx],
            rx,
            inner_iname=rx_inner,
            outer_iname=rx_outer,
        )
    if i_ry is not None:
        assert ry is not None
        t_unit = lp.split_iname(
            t_unit,
            free_indices[i_ry],
            ry,
            inner_iname=ry_inner,
            outer_iname=ry_outer,
        )

    # }}}

    # {{{ split redn indices

    for redn_iname, t_redn, inner_iname, outer_iname in szip(
        redn_indices, t_redns, inner_redn_inames, outer_redn_inames
    ):
        if t_redn != 1:
            t_unit = lp.split_iname(
                t_unit,
                redn_iname,
                t_redn,
                inner_iname=inner_iname,
                outer_iname=outer_iname,
            )

    # }}}

    precompute_outer_inames = frozenset(
        {
            redn_outer_iname if t_redn != 1 else redn_iname
            for redn_outer_iname, redn_iname, t_redn in szip(
                outer_redn_inames, redn_indices, t_redns
            )
        }
    )

    A_sweep_inames: list[str] = []
    A_prftch_inames: list[str | None] = []
    B_sweep_inames: list[str] = []
    B_prftch_inames: list[str | None] = []

    for (sweep_inames, prftch_inames), in_idx_set in szip(
        ((A_sweep_inames, A_prftch_inames), (B_sweep_inames, B_prftch_inames)),
        ensm.in_idx_sets,
    ):
        for in_idx in in_idx_set:
            if in_idx in ensm.out_idx_set:
                if ensm.out_idx_set.index(in_idx) == i_tx:
                    sweep_inames.append(tx_inner)
                    if tx != 1:
                        prftch_inames.append(vng("iprftch_tx"))
                    else:
                        prftch_inames.append(None)
                elif ensm.out_idx_set.index(in_idx) == i_ty:
                    sweep_inames.append(ty_inner)
                    if ty != 1:
                        prftch_inames.append(vng("iprftch_ty"))
                    else:
                        prftch_inames.append(None)
                elif ensm.out_idx_set.index(in_idx) == i_rx:
                    sweep_inames.append(rx_inner)
                    if rx != 1:
                        prftch_inames.append(vng("iprftch_rx"))
                    else:
                        prftch_inames.append(None)
                elif ensm.out_idx_set.index(in_idx) == i_ry:
                    sweep_inames.append(ry_inner)
                    if ry != 1:
                        prftch_inames.append(vng("iprftch_ry"))
                    else:
                        prftch_inames.append(None)
                else:
                    # Not sweep this iname as it is purely a precompute outer iname.
                    prftch_inames.append(None)
            else:
                assert in_idx in ensm_redn_indices
                if t_redns[ensm_redn_indices.index(in_idx)] == 1:
                    # Not sweep this iname as it is purely a precompute outer iname.
                    prftch_inames.append(None)
                else:
                    sweep_inames.append(
                        inner_redn_inames[ensm_redn_indices.index(in_idx)]
                    )
                    prftch_inames.append(vng(f"iprftch_{in_idx}"))

    # precompute A
    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        A,
        sweep_inames=A_sweep_inames,
        precompute_inames=A_prftch_inames,
        precompute_outer_inames=precompute_outer_inames,
        temporary_address_space=lp.AddressSpace.LOCAL,
        within=within,
        compute_insn_id=a_prftch_insn_id,
        default_tag=None,
    )

    assert (frozenset(A_prftch_inames) - {None} | precompute_outer_inames) == (
        t_unit[kernel_name].id_to_insn[a_prftch_insn_id].within_inames
    )
    logger.info("Precomputed A")

    # prefetch B
    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        B,
        sweep_inames=B_sweep_inames,
        precompute_inames=B_prftch_inames,
        precompute_outer_inames=precompute_outer_inames,
        temporary_address_space=lp.AddressSpace.LOCAL,
        compute_insn_id=b_prftch_insn_id,
        within=within,
        default_tag=None,
    )
    assert (frozenset(B_prftch_inames) - {None} | precompute_outer_inames) == (
        t_unit[kernel_name].id_to_insn[b_prftch_insn_id].within_inames
    )
    logger.info("Precomputed B")

    t_unit = lp.realize_reduction(t_unit, insn_id_filter=insn_id)
    logger.info("Realized reduction")

    (acc_name,) = (
        frozenset(t_unit[kernel_name].temporary_variables)
        & t_unit[kernel_name].id_to_insn[insn_id].read_dependency_names()
    )
    if i_rx is not None and i_ry is not None:
        inames_to_duplicate = sorted({rx_inner, ry_inner})
    elif i_rx is not None:
        inames_to_duplicate = sorted({rx_inner})
    elif i_ry is not None:
        inames_to_duplicate = sorted({ry_inner})
    else:
        inames_to_duplicate = []

    for iname_to_duplicate in inames_to_duplicate:
        t_unit = lp.privatize_temporaries_with_inames(
            t_unit, iname_to_duplicate, only_var_names={acc_name}
        )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames_to_duplicate,
        within=f"writes:{acc_name} and not reads:{acc_name}",
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        inames_to_duplicate,
        within=f"reads:{acc_name} and not writes:{acc_name}",
    )

    logger.info("Done with iname duplication")

    block_inames = [
        (
            tx_outer
            if ifree_idx == i_tx
            else (
                ty_outer
                if ifree_idx == i_ty
                else (
                    rx_outer
                    if ifree_idx == i_rx
                    else ry_outer if ifree_idx == i_ry else free_index
                )
            )
        )
        for ifree_idx, free_index in enumerate(free_indices)
    ]

    A_prftch_iname_x, A_prftch_iname_y = [
        iname for iname in A_prftch_inames[::-1] if iname
    ][:2]
    B_prftch_iname_x, B_prftch_iname_y = [
        iname for iname in B_prftch_inames[::-1] if iname
    ][:2]
    t_unit = lp.split_iname(t_unit, A_prftch_iname_x, tx, inner_tag="l.0")
    t_unit = lp.split_iname(t_unit, A_prftch_iname_y, ty, inner_tag="l.1")
    logger.info("Done with parallelizing prftch(A)")

    t_unit = lp.split_iname(t_unit, B_prftch_iname_x, tx, inner_tag="l.0")
    t_unit = lp.split_iname(t_unit, B_prftch_iname_y, ty, inner_tag="l.1")
    logger.info("Done with parallelizing prftch(B)")

    t_unit = lp.join_inames(t_unit, block_inames, iouter)
    logger.info("Done with join inames")

    t_unit = lp.tag_inames(t_unit, {iouter: "g.0"})
    t_unit = lp.add_inames_to_insn(
        t_unit,
        frozenset([iouter]),
        lp_match.Or((lp_match.Id(a_prftch_insn_id), lp_match.Id(b_prftch_insn_id))),
    )

    return t_unit


if __name__ == "__main__":
    import os

    import pyopencl as cl

    cl_ctx = cl.create_some_context()

    # expr = fnsm.einsum("aebf,fdec->abcd",
    #                    fnsm.array((72, 72, 72, 72), np.float64),
    #                    fnsm.array((72, 72, 72, 72), np.float64),
    #                    arg_names=["A", "B"])

    expr = fnsm.einsum(
        "il,ljk->ijk",
        fnsm.array("A", (72, 72), np.float64),
        fnsm.array("B", (72, 72, 72), np.float64),
    )

    fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx)



import logging
import math
from typing import TYPE_CHECKING, Any, cast

import islpy as isl
import loopy as lp
import loopy.match as lp_match
from more_itertools import chunked
from more_itertools import zip_equal as szip

import feinsum as fnsm
import feinsum.loopy_utils as lp_utils
from feinsum.tuning import IntParameter

if TYPE_CHECKING:
    from collections.abc import Collection

logger = logging.getLogger(__name__)


@fnsm.tuning.einsum_arg("noutputs", lambda e: e.b)
@fnsm.tuning.einsum_arg("ndim", lambda e: e.shape[0])
@fnsm.tuning.einsum_arg("ndof", lambda e: e.shape[2])
@fnsm.tuning.transform_param("n_e_per_wg", lambda e: IntParameter(2, 32))
@fnsm.tuning.transform_param(
    "nwork_items_per_e", lambda e: IntParameter(1, e.shape[2])
)
@fnsm.tuning.transform_param(
    "n_stmt_tile", lambda e: IntParameter(1, math.ceil(e.b))
)
@fnsm.tuning.transform_param(
    "j_tiles", lambda e: IntParameter(1, math.ceil(e.shape[2] / 2))
)
@fnsm.tuning.transform_param(
    "i_tiles", lambda e: IntParameter(1, math.ceil(e.shape[2] / 2))
)
def transform(
    t_unit: lp.TranslationUnit,
    noutputs: int,
    ndim: int,
    ndof: int,
    n_e_per_wg: int,
    nwork_items_per_e: int,
    n_stmt_tile: int,
    i_tiles: int,
    j_tiles: int,
    # FIXME: Making this is BoolParameters leads to an error in validation.
    prftch_u_to_local: bool = False,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:

    if n_e_per_wg * nwork_items_per_e > 600:
        raise fnsm.InvalidParameterError("Block dimension limit exceeded")

    if (
        (ndim * math.ceil((ndof) / i_tiles) * math.ceil(ndof / j_tiles))
        + int(prftch_u_to_local)
        * ndof
        * n_e_per_wg
        * math.ceil(noutputs / n_stmt_tile)
    ) * 8e-3 > 47:
        raise fnsm.InvalidParameterError("Shared memory limit exceeded")

    from loopy.match import parse_match
    from pymbolic import variables

    kernel_name = kernel_name or t_unit.default_entrypoint.name

    within = parse_match(insn_match)

    ref_einsum = fnsm.batched_einsum(
        "xre,rij,ej->xei",
        [
            [
                fnsm.array("J", (ndim, ndim, "Nel")),
                fnsm.array("D", (ndim, ndof, ndof)),
                fnsm.array(f"u{i}", ("Nel", ndof)),
            ]
            for i in range(noutputs)
        ],
    )

    # {{{ get corresponding variables in t_unit

    vng = t_unit.default_entrypoint.get_var_name_generator()
    ing = t_unit.default_entrypoint.get_instruction_id_generator()
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    i = subst_map["i"]
    j = subst_map["j"]
    J = subst_map["J"]
    e = subst_map["e"]
    D = subst_map["D"]
    x = subst_map["x"]
    r = subst_map["r"]
    e_inner, e_outer = f"{e}_inner", f"{e}_outer"
    us = tuple(subst_map[f"u{i}"] for i in range(noutputs))
    ref_outputs = ["_fe_out"] + [
        f"_fe_out_{ioutput}" for ioutput in range(noutputs - 1)
    ]
    outputs = tuple(subst_map[ref_output] for ref_output in ref_outputs)
    del ref_outputs
    # prftch_J = ing(f"prftch_{J}")

    # }}}

    # {{{ get new names

    n_stmt_tile = math.ceil(noutputs / math.ceil(noutputs / n_stmt_tile))

    new_is = [vng("i") for _ in range(n_stmt_tile)]
    new_js = [vng("j") for _ in range(n_stmt_tile)]
    new_xs = [vng("x") for _ in range(n_stmt_tile)]
    new_rs = [vng("r") for _ in range(n_stmt_tile)]
    new_es = [vng("e") for _ in range(n_stmt_tile)]
    i_stmt_tile_to_fields = tuple(
        list(el) for el in chunked(us, math.ceil(noutputs / n_stmt_tile))
    )
    i_stmt_tile_to_outputs = tuple(
        list(el) for el in chunked(outputs, math.ceil(noutputs / n_stmt_tile))
    )

    ufetch_ids = tuple(
        tuple(ing("u_fetch_insn") for _ in fields)
        for fields in i_stmt_tile_to_fields
    )

    # }}}

    for new_i, new_j, new_x, new_r, new_e, outputs_in_tile in szip(
        new_is, new_js, new_xs, new_rs, new_es, i_stmt_tile_to_outputs
    ):
        t_unit = lp.duplicate_inames(
            t_unit,
            [i, j, x, r, e],
            within=lp_match.And(
                (
                    within,
                    lp_match.Or(
                        tuple(lp_match.Writes(output) for output in outputs_in_tile)
                    ),
                )
            ),
            new_inames=[new_i, new_j, new_x, new_r, new_e],
        )
        t_unit = t_unit.with_kernel(
            lp_utils.decouple_domain(
                t_unit[kernel_name],
                [new_i, new_j, new_x, new_r, new_e],
                parent_inames=cast(
                    "Collection[str]",
                    t_unit[kernel_name]
                    .get_inames_domain(e)
                    .get_var_names(isl.dim_type.param),
                ),
            )
        )

    t_unit = lp.remove_unused_inames(t_unit)
    D_fetch = vng(f"{D}_fetch")
    D_fetch_ids = tuple(ing("D_fetch_id") for _ in range(n_stmt_tile))

    if 0:
        # TODO: Assumes that len(i_inner_inner) > len(jprftch_D_inner).
        # Re-enable this after that ambiguity is fixed.
        J_fetch = vng(f"{J}_fetch")
        J_prftch_0, J_prftch_1 = vng(f"{J}_prftch_x"), vng(f"{J}_prftch_r")
        t_unit = lp.precompute(
            t_unit,
            J,
            sweep_inames=[x, r],
            precompute_outer_inames=frozenset({}),
            temporary_address_space=lp.AddressSpace.PRIVATE,
            temporary_name=J_fetch,
            precompute_inames=(J_prftch_0, J_prftch_1),
            default_tag=None,
            within=within,
        )

    for istmt_tile in range(n_stmt_tile):

        i = new_is[istmt_tile]
        j = new_js[istmt_tile]
        x = new_xs[istmt_tile]
        r = new_rs[istmt_tile]
        e = new_es[istmt_tile]

        # {{{ set var names

        j_inner = vng(f"{j}_inner")
        j_tile = f"{j}_tile"
        i_tile, i_inner = f"{i}_tile", f"{i}_inner"
        e_prcmpt_subst = vng(f"{e}_prcmpt")
        r_prcmpt_subst = vng(f"{r}_prcmpt")
        i_prcmpt_subst = vng(f"{i}_prcmpt")
        rprftch_D, iprftch_D, jprftch_D = (
            vng("rprftchD"),
            vng("iprftchD"),
            vng("jprftchD"),
        )

        prcmpt_j_redns = tuple(
            ing(f"prcmpt_{j}_redn") for _ in i_stmt_tile_to_fields[istmt_tile]
        )
        i_inner_inner, i_inner_outer = (
            vng(f"{i_inner}_inner"),
            vng(f"{i_inner}_outer"),
        )

        subst_names = tuple(vng("subst") for _ in i_stmt_tile_to_fields[istmt_tile])

        # }}}

        # {{{ term hoisting to match the flop count of opt_einsum

        knl = t_unit[kernel_name]
        knl = lp.split_reduction_inward(knl, j)
        knl = lp_utils.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, j)

        for subst_name, u, output in szip(
            subst_names,
            i_stmt_tile_to_fields[istmt_tile],
            i_stmt_tile_to_outputs[istmt_tile],
        ):
            knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
                knl,
                subst_name=subst_name,
                arguments=variables(f"{e} {i} {r}"),
                within=lp_match.And((within, lp_match.Writes(output))),
                terms_filter=lambda x: fnsm.get_call_ids(x) <= {D, u},
            )

        t_unit = t_unit.with_kernel(knl)

        t_unit = lp.split_iname(
            t_unit,
            i,
            math.ceil(ndof / i_tiles),
            outer_iname=i_tile,
            inner_iname=i_inner,
        )

        # }}}

        t_unit = lp.split_iname(
            t_unit,
            e,
            n_e_per_wg,
            inner_iname=e_inner,
            outer_iname=e_outer,
            inner_tag="l.1",
            outer_tag="g.0",
        )

        # {{{ prefetch 'u'

        if prftch_u_to_local:
            raise NotImplementedError
        else:
            jprftch_u = vng("j_prftch_u")
            for u, ufetch_id in zip(
                i_stmt_tile_to_fields[istmt_tile],
                ufetch_ids[istmt_tile],
                strict=False,
            ):
                t_unit = lp.precompute(  # type: ignore[no-untyped-call]
                    t_unit,
                    u,
                    sweep_inames=[j],
                    precompute_outer_inames=frozenset([e_inner, e_outer]),
                    temporary_address_space=lp.AddressSpace.PRIVATE,
                    temporary_name=vng("u_prftch_var"),
                    default_tag="unr",
                    precompute_inames=(
                        None,
                        jprftch_u,
                    ),
                    compute_insn_id=ufetch_id,
                    within=within,
                )

        # }}}

        # {{{ tile and prefetch D

        t_unit = lp.split_iname(
            t_unit,
            j,
            math.ceil(ndof / j_tiles),
            inner_iname=j_inner,
            outer_iname=j_tile,
            inner_tag="unr",
            outer_tag="unr",
        )
        t_unit = lp.precompute(  # type: ignore[no-untyped-call]
            t_unit,
            D,
            [i_inner, r, j_inner],
            precompute_outer_inames=frozenset([e_outer, i_tile, j_tile]),
            precompute_inames=[rprftch_D, iprftch_D, jprftch_D],
            temporary_address_space=lp.AddressSpace.LOCAL,
            temporary_name=D_fetch,
            compute_insn_id=D_fetch_ids[istmt_tile],
            within=lp_match.And(
                (
                    within,
                    lp_match.Or(
                        tuple(
                            lp_match.Writes(output)
                            for output in i_stmt_tile_to_outputs[istmt_tile]
                        )
                    ),
                )
            ),
            default_tag=None,
        )
        t_unit = lp.split_iname(t_unit, iprftch_D, n_e_per_wg, inner_tag="l.1")
        t_unit = lp.split_iname(
            t_unit, jprftch_D, nwork_items_per_e, inner_tag="l.0"
        )

        # }}}

        t_unit = lp.split_iname(
            t_unit,
            i_inner,
            nwork_items_per_e,
            inner_iname=i_inner_inner,
            outer_iname=i_inner_outer,
            inner_tag="l.0",
            outer_tag="unr",
        )

        # {{{ precompute 'subst'

        for subst_name, prcmpt_j_redn in szip(subst_names, prcmpt_j_redns):
            t_unit = lp.precompute(  # type: ignore[no-untyped-call]
                t_unit,
                subst_name,
                sweep_inames=[r, i_inner_outer],
                precompute_inames=[e_prcmpt_subst, i_prcmpt_subst, r_prcmpt_subst],
                # storage_axes=[0, 1],
                precompute_outer_inames=frozenset(
                    {e_inner, e_outer, i_tile, i_inner_inner}
                ),
                default_tag="unr",
                compute_insn_id=prcmpt_j_redn,
                temporary_name=vng("tmp_hoist_j_redn"),
                temporary_address_space=lp.AddressSpace.PRIVATE,
            )

        # }}}

        # {{{ TODO: remove once github.com/inducer/loopy/issues/666 is resolved.

        t_unit = lp.realize_reduction(
            t_unit, insn_id_filter=frozenset(prcmpt_j_redns)
        )
        inames_to_duplicate = (
            frozenset({i_prcmpt_subst, r_prcmpt_subst})
            & t_unit[kernel_name].all_inames()
        )
        acc_names = set()
        for prcmpt_j_redn in prcmpt_j_redns:
            (acc_name,) = t_unit[kernel_name].id_to_insn[
                prcmpt_j_redn
            ].read_dependency_names() & set(t_unit[kernel_name].temporary_variables)
            assert acc_name.startswith("acc")
            acc_names.add(acc_name)

        t_unit = lp.privatize_temporaries_with_inames(
            t_unit, inames_to_duplicate, only_var_names=acc_names
        )

        t_unit = lp.duplicate_inames(
            t_unit,
            inames_to_duplicate,
            within=lp_match.Or(
                tuple(
                    lp_match.And(
                        (
                            lp_match.Writes(acc_name),
                            lp_match.Not(lp_match.Reads(acc_name)),
                        )
                    )
                    for acc_name in acc_names
                )
            ),
        )
        t_unit = lp.duplicate_inames(
            t_unit,
            inames_to_duplicate,
            within=lp_match.Or(
                tuple(
                    lp_match.And(
                        (
                            lp_match.Reads(acc_name),
                            lp_match.Not(lp_match.Writes(acc_name)),
                        )
                    )
                    for acc_name in acc_names
                )
            ),
        )

        # }}}

        if not prftch_u_to_local:
            # TODO: Yet another headache to ensure that the fetch instruction uses
            # all the hw axes.
            for ufetch_id in ufetch_ids[istmt_tile]:
                t_unit = lp.add_inames_to_insn(
                    t_unit, i_inner_inner, lp_match.Id(ufetch_id)
                )

    if prftch_u_to_local:
        # add dependencies correctly.
        raise NotImplementedError

    for istmt_tile in range(1, n_stmt_tile):
        preceding_outputs = i_stmt_tile_to_outputs[istmt_tile - 1]
        t_unit = lp.add_dependency(
            t_unit,
            insn_match=lp_match.Or(
                (
                    lp_match.Id(D_fetch_ids[istmt_tile]),
                    lp_match.Or(
                        tuple(
                            lp_match.Id(ufetch_id)
                            for ufetch_id in ufetch_ids[istmt_tile]
                        )
                    ),
                )
            ),
            depends_on=lp_match.And(
                (
                    within,
                    lp_match.Or(
                        tuple(
                            lp_match.Writes(output) for output in preceding_outputs
                        )
                    ),
                )
            ),
        )

    return t_unit


if __name__ == "__main__":
    import os
    from functools import partial

    import pyopencl as cl

    Ndim = 3
    Ndof = 4
    Nfields = 5

    cl_ctx = cl.create_some_context()

    expr = fnsm.batched_einsum(
        "xre,rij,ej->xei",
        [
            [
                fnsm.array("J", (Ndim, Ndim, "Nel")),
                fnsm.array("D", (Ndim, Ndof, Ndof)),
                fnsm.array(f"u{i}", ("Nel", Ndof)),
            ]
            for i in range(Nfields)
        ],
    )

    if 1:
        fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx)
    else:
        # Enable while debugging ->
        # evaluate a point in the parameter space.
        bound_transform = partial(
            transform,
            ndim=Ndim,
            ndof=Ndof,
            noutputs=Nfields,
            n_e_per_wg=21,
            nwork_items_per_e=12,
            n_stmt_tile=Nfields,
            i_tiles=3,
            j_tiles=1,
            prftch_u_to_local=False,
        )

        print(
            fnsm.stringify_comparison_vs_roofline(
                expr, transform=bound_transform, cl_ctx=cl_ctx
            )
        )

# vim: fdm=marker



import os

import pyopencl as cl

import feinsum as fnsm

NEW_DB_NAME = "transform_archive_v5.db"
TOP_K = 10  # how many entries to reevaluate


def main():
    from feinsum.sql_utils import get_timed_einsums_in_db, record_into_db
    from feinsum.tuning import _get_impls_path

    cl_ctx = cl.create_some_context()
    cq = cl.CommandQueue(cl_ctx)

    for i, einsum in enumerate(get_timed_einsums_in_db(cq.device)):
        best_queries = sorted(
            fnsm.query(einsum, cl_ctx), key=lambda q: -q.giga_op_rate("float64")
        )[:TOP_K]
        if os.path.exists(NEW_DB_NAME):
            recorded_queries = fnsm.query(einsum, cl_ctx, database=NEW_DB_NAME)
            recorded_transform_params = [
                q.transform_params for q in recorded_queries
            ]
        else:
            recorded_transform_params = []

        print(i, len(best_queries))
        for iquery, query in enumerate(best_queries):
            if query.transform_params in recorded_transform_params:
                print("Skipping as completed")
                continue
            print(
                f"({iquery}, {query.transform_id}):"
                f" Old GFLOPS/s = {query.giga_op_rate('float64')}."
            )
            print(query.transform_params)
            record_into_db(
                einsum,
                cl_ctx,
                module_path=os.path.join(_get_impls_path(), query.transform_id),
                transform_params=query.transform_params,
                database=NEW_DB_NAME,
            )
            print(75 * "=")


if __name__ == "__main__":
    main()



# NOTE: Keep this implementation short as this is a part of docs.
# Type-ignoring because this is part of the demo
# type: ignore
# flake8: noqa

# BEGINEXAMPLE
import feinsum as f
import loopy as lp
import numpy as np

from feinsum.tuning import IntParameter, einsum_arg, transform_param


@einsum_arg("N", lambda ensm: ensm.arg_shapes[0][1])
@transform_param("n_x_per_wg", lambda ensm: IntParameter(2, 32))
@transform_param("n_wi_per_x", lambda ensm: IntParameter(1, ensm.arg_shapes[0][1]))
def transform(t_unit, N, n_x_per_wg, n_wi_per_x, insn_match=None, kernel_name=None):
    ref_einsum = f.einsum(
        "xik,kj->xij",
        f.array((np.inf, N, N), np.float64),
        f.array((N, N), np.float64),
        arg_names=["A", "B"],
    )

    subst_map = f.match_t_unit_to_einsum(t_unit, ref_einsum, insn_match=insn_match)

    vng = t_unit.default_entrypoint.get_var_name_generator()
    x = subst_map["x"]
    i = subst_map["i"]
    j = subst_map["j"]
    k = subst_map["k"]
    B = subst_map["B"]

    Bprftch_k, Bprftch_j = vng("Bprftch_k"), vng("Bprftch_j")
    x_outer = vng(f"{x}_outer")

    # Notation:
    #    - The einsum computes 'out[x, i, j]'.
    #
    # Transform space:
    #   - Each work-group computes the slice:
    #     'out[group_id(0)*n_x_per_wg:(group_id(0)+1)*n_x_per_wg,:,:]'
    #   - Each work_item computes the slice:
    #     'out[group_id(0)+n_x_per_wg*local_id(1),local_id(0)::nwi_per_x,:]'

    # Loop Transformations: Work Division
    # -----------------------------------
    t_unit = lp.split_iname(
        t_unit, x, n_x_per_wg, outer_iname=x_outer, inner_tag="l.1", outer_tag="g.0"
    )

    t_unit = lp.split_iname(t_unit, i, n_wi_per_x, inner_tag="l.0")

    # Data Transformations: Precompute 'B'
    # ------------------------------------
    t_unit = lp.precompute(
        t_unit,
        B,
        sweep_inames={k, j},
        precompute_outer_inames=frozenset([x_outer]),
        precompute_inames=[Bprftch_k, Bprftch_j],
        default_tag="l.auto",
    )

    return t_unit


# ENDEXAMPLE


if __name__ == "__main__":
    import pyopencl as cl
    import os

    cl_ctx = cl.create_some_context()

    my_einsum = f.einsum(
        "xik,kj->xij",
        f.array((np.inf, 35, 35), np.float64),
        f.array((35, 35), np.float64),
    )
    f.autotune(
        my_einsum,
        os.path.abspath(__file__),
        cl_ctx,
        long_dim_length=1_000,
        stop_after=5,
    )



from typing import Any

import loopy as lp
import loopy.match as lp_match

import feinsum as fnsm
from feinsum.tuning import IntParameter


@fnsm.tuning.einsum_arg("ndofs", lambda e: e.shape[1])
@fnsm.tuning.transform_param(
    "nworkitems_per_e", lambda e: IntParameter(1, e.shape[1])
)
@fnsm.tuning.transform_param("n_e_per_wg", lambda e: IntParameter(1, 32))
def transform(
    t_unit: lp.TranslationUnit,
    ndofs: int,
    n_e_per_wg: int,
    nworkitems_per_e: int,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:

    kernel_name = kernel_name or t_unit.default_entrypoint.name
    within = lp_match.parse_match(insn_match)
    (insn_id,) = [
        insn.id
        for insn in t_unit[kernel_name].instructions
        if within(t_unit[kernel_name], insn)
    ]

    ref_einsum = fnsm.einsum(
        "ij,ej->ei", fnsm.array("D", (ndofs, ndofs)), fnsm.array("u", ("Nel", ndofs))
    )
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, kernel_name=kernel_name, insn_match=insn_match
    )
    vng = t_unit[kernel_name].get_var_name_generator()
    ing = t_unit[kernel_name].get_instruction_id_generator()

    e = subst_map["e"]
    i = subst_map["i"]
    j = subst_map["j"]
    i_inner, i_outer = vng(f"{i}_inner"), vng(f"{i}_outer")
    e_outer = vng(f"{e}_outer")
    # out = subst_map["_fe_out"]
    u = subst_map["u"]
    D = subst_map["D"]
    uprftch_insn_id = ing("u_prftch_id")
    Dprftch_insn_id = ing("D_prftch_id")
    jprftch_u = vng("jprftch_u")
    iprftch_D, jprftch_D = vng("iprftch_D"), vng("jprftch_D")

    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        u,
        sweep_inames=[j],
        precompute_outer_inames=frozenset([e]),
        compute_insn_id=uprftch_insn_id,
        precompute_inames=(
            None,
            jprftch_u,
        ),
        default_tag="unr",
        temporary_address_space=lp.AddressSpace.PRIVATE,
        within=within,
    )

    t_unit = lp.split_iname(
        t_unit, e, n_e_per_wg, outer_iname=e_outer, inner_tag="l.1", outer_tag="g.0"
    )

    # {{{ prefetch "D"

    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        D,
        sweep_inames=[i, j],
        precompute_outer_inames=frozenset([e_outer]),
        compute_insn_id=Dprftch_insn_id,
        precompute_inames=(iprftch_D, jprftch_D),
        default_tag=None,
        temporary_address_space=lp.AddressSpace.LOCAL,
        within=within,
    )
    t_unit = lp.split_iname(t_unit, iprftch_D, nworkitems_per_e, inner_tag="l.0")
    t_unit = lp.split_iname(t_unit, jprftch_D, n_e_per_wg, inner_tag="l.1")

    # }}}

    t_unit = lp.split_iname(
        t_unit,
        i,
        nworkitems_per_e,
        outer_iname=i_outer,
        inner_iname=i_inner,
        inner_tag="l.0",
        outer_tag="unr",
    )

    t_unit = lp.add_inames_to_insn(
        t_unit, frozenset([i_inner]), insn_match=lp_match.Id(uprftch_insn_id)
    )
    t_unit = lp.realize_reduction(t_unit, insn_id_filter=frozenset([insn_id]))
    (acc_name,) = t_unit[kernel_name].id_to_insn[
        insn_id
    ].read_dependency_names() & set(t_unit[kernel_name].temporary_variables)
    inames_to_duplicate = frozenset([i_outer])
    assert acc_name.startswith("acc")
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, inames_to_duplicate, only_var_names={acc_name}
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        inames_to_duplicate,
        within=f"writes:{acc_name} and not reads:{acc_name}",
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        inames_to_duplicate,
        within=f"reads:{acc_name} and not writes:{acc_name}",
    )
    t_unit = lp.tag_inames(t_unit, {j: "unr"})

    return t_unit


if __name__ == "__main__":
    import os

    import pyopencl as cl

    Ndof = 15
    Nfields = 1

    cl_ctx = cl.create_some_context()

    expr = fnsm.einsum(
        "ij,ej->ei", fnsm.array("D", (Ndof, Ndof)), fnsm.array("u", ("Nel", Ndof))
    )

    fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx)

# vim: fdm=marker



"""
.. currentmodule:: feinsum.measure

.. autofunction:: timeit
.. autofunction:: measure_giga_op_rate
.. autofunction:: stringify_comparison_vs_roofline
"""

import logging
from collections.abc import Mapping
from typing import Any, cast

import loopy as lp
import numpy as np
import numpy.typing as npt
import pymbolic.primitives as prim
import pyopencl as cl
import pyopencl.array as cla
from immutables import Map

from feinsum.diagnostics import NoDevicePeaksInfoError
from feinsum.einsum import (
    INT_CLASSES,
    BatchedEinsum,
    ContractionSchedule,
    IntegralT,
)
from feinsum.typing import ToStr, TransformT

logger = logging.getLogger(__name__)


N_WARMUP_ROUNDS = 20
N_MIN_TIMING_ROUNDS = 500
N_MIN_SIM_SECS = 2


def get_real_dtype(dtype: np.dtype[Any]) -> np.dtype[Any]:
    return np.empty(0, dtype=dtype).real.dtype


def generate_out_arrays(
    queue: cl.CommandQueue, t_unit: lp.TranslationUnit
) -> Map[str, cla.Array]:
    t_unit = lp.preprocess_kernel(t_unit)
    knl = t_unit.default_entrypoint

    out_buffers = {}

    for arg in knl.args:
        if arg.is_output:
            assert all(isinstance(dim, INT_CLASSES) for dim in arg.shape)
            assert arg.dtype is not None
            out_buffers[arg.name] = cla.zeros(
                queue, shape=arg.shape, dtype=arg.dtype.numpy_dtype
            )

    return Map(out_buffers)


def _generate_random_np_array(
    rng: "np.random._generator.Generator",
    dtype: np.dtype[Any],
    shape: tuple[IntegralT, ...],
) -> npt.NDArray[Any]:
    if dtype.kind == "c":
        real_dtype = get_real_dtype(dtype)
        # type-ignored because numpy addition types not quite precise
        return rng.random(  # type: ignore[no-any-return]
            size=shape, dtype=real_dtype
        ) + dtype.type(1j) * rng.random(size=shape, dtype=real_dtype)
    elif dtype.kind == "i":
        return rng.integers(low=-100, high=100, size=shape, dtype=dtype)
    else:
        return rng.random(size=shape, dtype=dtype)


def generate_input_arrays(
    queue: cl.CommandQueue,
    einsum: BatchedEinsum,
    long_dim_length: int,
    np_seed: int = 0,
) -> Map[str, cla.Array]:
    from numpy.random import default_rng

    # {{{ compute val_to_shape

    val_to_shape: dict[str, tuple[IntegralT, ...]] = {}

    for arg, shape in einsum.arg_to_shape.items():
        val_to_shape[arg] = tuple(
            dim if isinstance(dim, INT_CLASSES) else long_dim_length for dim in shape
        )

    # }}}

    rng = default_rng(np_seed)

    return Map(
        {
            name: cla.to_device(
                queue, _generate_random_np_array(rng, dtype, val_to_shape[name])
            )
            for name, dtype in einsum.arg_to_dtype.items()
        }
    )


def validate_batched_einsum_transform(
    einsum: BatchedEinsum,
    cl_ctx: cl.Context,
    transform: TransformT,
    schedule: ContractionSchedule | None = None,
) -> None:
    """
    If the :class:`loopy.LoopKernel` generated from *einsum* does not replicate
    the results after being transformed with *transform*, then a
    :class:`RuntimeError` is raised.
    """

    from feinsum.codegen.loopy import generate_loopy

    cq = cl.CommandQueue(cl_ctx)

    ref_t_unit = generate_loopy(einsum, schedule=schedule)
    ref_t_unit = lp.set_options(ref_t_unit, no_numpy=True, return_dict=True)
    long_dim_length = 100

    arg_dict: dict[str, cla.Array | int] = dict(
        generate_input_arrays(cq, einsum, long_dim_length)
    )
    arg_dict.update(
        dict.fromkeys(ref_t_unit.default_entrypoint.all_params(), long_dim_length)
    )

    t_unit = transform(ref_t_unit, insn_match=None, kernel_name=None)

    # {{{ get output buffers

    ref_outs = generate_out_arrays(
        cq,
        lp.fix_parameters(
            t_unit,
            within=None,
            **dict.fromkeys(t_unit.default_entrypoint.all_params(), long_dim_length),
        ),
    )
    transform_outs = Map(
        {name: cla.zeros_like(ary) for name, ary in ref_outs.items()}
    )

    # }}}

    # pylint-disable-reason: for some reason pylint thinks ref_t_unit is not callable
    evt, ref_outs = ref_t_unit(
        cq, **arg_dict, **ref_outs  # pylint: disable=not-callable
    )
    evt.wait()
    evt, transform_outs = t_unit(cq, **arg_dict, **transform_outs)
    evt.wait()

    if frozenset(ref_outs.keys()) != frozenset(transform_outs.keys()):
        raise RuntimeError("Output names mismatch")

    for name, ref_out in sorted(ref_outs.items()):
        ref_out_np = ref_out.get()
        transform_out_np = transform_outs[name].get()
        dtype = ref_out_np.dtype
        if dtype != transform_out_np.dtype:
            raise RuntimeError(f"dtype mismatch for output '{name}'")

        real_dtype = get_real_dtype(dtype)

        if real_dtype == np.float32:
            atol = 1e-6
            rtol = 1e-6
        elif real_dtype == np.float64:
            atol = 1e-14
            rtol = 1e-14
        else:
            raise NotImplementedError(real_dtype)

        np.testing.assert_allclose(
            transform_out_np, ref_out_np, atol=atol, rtol=rtol
        )

    logger.info("Statistically verified the soundness of the transformation")


def timeit(
    einsum: BatchedEinsum,
    *,
    transform: TransformT,
    cl_ctx: cl.Context,
    long_dim_length: int = 100000,
    schedule: ContractionSchedule | None = None,
) -> float:
    """
    Returns the runtime in seconds for executing *einsum* on OpenCL context
    *cl_ctx*.

    :param transform: The transformation to be applied to
        :class:`loopy.TranslationUnit` lowered from *einsum*.
    """
    from time import time

    from feinsum.codegen.loopy import generate_loopy

    # Validate the transformation before fusing it
    validate_batched_einsum_transform(einsum, cl_ctx, transform, schedule)

    cq = cl.CommandQueue(cl_ctx)

    t_unit = generate_loopy(einsum, schedule=schedule)
    t_unit = lp.set_options(t_unit, no_numpy=True, return_dict=True)

    param_dict = generate_input_arrays(cq, einsum, long_dim_length)
    out_dict = generate_out_arrays(
        cq,
        lp.fix_parameters(
            t_unit,
            within=None,
            **dict.fromkeys(t_unit.default_entrypoint.all_params(), long_dim_length),
        ),
    )

    t_unit = transform(t_unit, insn_match=None, kernel_name=None)

    arg_dict = param_dict.update(out_dict)

    # {{{ WARMUP

    for _ in range(N_WARMUP_ROUNDS):
        evt, _ = t_unit(cq, **arg_dict)

    cq.finish()

    # }}}

    total_sim_time = 0.0
    total_rounds = 0

    while (total_rounds < N_MIN_TIMING_ROUNDS) or (total_sim_time < N_MIN_SIM_SECS):

        evt.wait()

        clock_start = time()

        for _ in range(10):
            evt, _ = t_unit(cq, **arg_dict)

        evt.wait()
        clock_end = time()

        total_sim_time += clock_end - clock_start
        total_rounds += 10

    return total_sim_time / total_rounds


def _get_giga_ops_from_einsum(
    expr: BatchedEinsum,
) -> Map[np.dtype[Any], prim.Expression]:
    from loopy.symbolic import qpolynomial_to_expr

    from feinsum.codegen.loopy import generate_loopy_with_opt_einsum_schedule

    t_unit = generate_loopy_with_opt_einsum_schedule(
        expr, use_blas=False, optimize="optimal"
    )

    kernel = t_unit.default_entrypoint
    kernel = kernel.copy(
        silenced_warnings=(
            [
                *kernel.silenced_warnings,
                "insn_count_subgroups_upper_bound",
                "summing_if_branches_ops",
            ]
        )
    )
    t_unit = t_unit.with_kernel(kernel)
    op_map = lp.get_op_map(t_unit, subgroup_size=1)
    new_op_map: dict[np.dtype[Any], prim.Expression] = {}

    for dtype in {op.dtype.numpy_dtype for op in op_map.keys()}:
        if dtype.kind == "c":
            c_ops = {
                op_type: op_map.filter_by(
                    dtype=[dtype], name=op_type, kernel_name=kernel.name
                )
                for op_type in ["add", "mul", "div"]
            }

            pwqpoly = (
                2 * c_ops["add"].sum()
                + 6 * c_ops["mul"].sum()
                + (6 + 3 + 2) * c_ops["div"].sum()
            ).pwqpolynomial
            dtype = get_real_dtype(dtype)
        else:
            pwqpoly = (
                op_map.filter_by(dtype=[dtype], kernel_name=kernel.name)
                .sum()
                .pwqpolynomial
            )

        new_op_map.setdefault(dtype, 0)

        if pwqpoly.n_piece() > 0:
            ((_, qpoly),) = pwqpoly.get_pieces()
            new_op_map[dtype] = new_op_map[dtype] + qpolynomial_to_expr(qpoly) * 1e-9  # type: ignore[no-untyped-call]

    return Map(new_op_map)


def _get_footprint_gbytes(expr: BatchedEinsum, long_dim_length: int) -> float:
    from feinsum.codegen.loopy import generate_loopy

    t_unit = generate_loopy(expr)
    t_unit = lp.fix_parameters(
        t_unit,
        within=None,
        **dict.fromkeys(t_unit.default_entrypoint.all_params(), long_dim_length),
    )
    t_unit = lp.infer_unknown_types(t_unit)
    kernel = t_unit.default_entrypoint

    result = (
        sum(
            np.prod(arg.shape) * cast("lp.LoopyType", arg.dtype).itemsize
            for arg in kernel.args
        )
        * 1e-9
    )
    assert isinstance(result, float)
    return result


def measure_giga_op_rate(
    expr: BatchedEinsum,
    *,
    transform: TransformT,
    cl_ctx: cl.Context,
    long_dim_length: int = 100000,
    schedule: ContractionSchedule | None = None,
) -> Map[np.dtype[Any], float]:
    """
    Returns the arithmetic operations rate (in Giga Ops per second) by
    arithmetic operation's result dtypes.
    """
    runtime = timeit(
        expr,
        transform=transform,
        cl_ctx=cl_ctx,
        long_dim_length=long_dim_length,
        schedule=schedule,
    )

    from pymbolic.mapper.evaluator import evaluate_to_float

    eval_context = {param.name: long_dim_length for param in expr.all_size_params}
    return Map(
        {
            k: evaluate_to_float(v, eval_context) / runtime
            for k, v in _get_giga_ops_from_einsum(expr).items()
        }
    )


def get_roofline_flop_rate(
    expr: BatchedEinsum, dev_name: str, long_dim_length: int = 100_000
) -> Map[np.dtype[Any], float]:

    from pymbolic.mapper.evaluator import evaluate_to_float

    from feinsum.data.device_info import DEV_TO_PEAK_BW, DEV_TO_PEAK_GFLOPS

    dtype_to_gflops_expr = _get_giga_ops_from_einsum(expr)
    ngbs = _get_footprint_gbytes(expr, long_dim_length)

    dtype_to_gflops = {
        dtype: evaluate_to_float(
            giga_ops_aff,
            {param.name: long_dim_length for param in expr.all_size_params},
        )
        for dtype, giga_ops_aff in dtype_to_gflops_expr.items()
    }
    try:
        roofline_time_due_to_flops = max(
            ngflops / DEV_TO_PEAK_GFLOPS[dev_name][dtype.name]
            for dtype, ngflops in dtype_to_gflops.items()
        )
        roofline_time_due_to_global_bw = ngbs / DEV_TO_PEAK_BW[dev_name]
    except KeyError as exc:
        raise NoDevicePeaksInfoError from exc
    roofline_time = max(roofline_time_due_to_flops, roofline_time_due_to_global_bw)

    return Map(
        {dtype: gflops / roofline_time for dtype, gflops in dtype_to_gflops.items()}
    )


def _strify_measured_vs_roofline(
    measured_flop_rate: Mapping[np.dtype[Any], ToStr],
    roofline_flop_rate: Mapping[np.dtype[Any], ToStr],
) -> str:
    try:
        from tabulate import tabulate
    except ImportError as exc:
        raise ImportError(
            "`tabulate` is need for pretty printing."
            " Install via `pip install tabulate`."
        ) from exc
    assert set(measured_flop_rate.keys()) == set(roofline_flop_rate.keys())
    perf_table = [["Dtype", "Measured GOps/s", "Roofline GOps/s"]]
    for dtype in sorted(measured_flop_rate.keys(), key=lambda x: x.itemsize):
        measured_flops = (
            f"{measured_flop_rate[dtype]:.1f}"
            if isinstance(measured_flop_rate[dtype], float)
            else str(measured_flop_rate[dtype])
        )
        roofline_flops = (
            f"{(roofline_flop_rate[dtype]):.1f}"
            if isinstance(roofline_flop_rate[dtype], float)
            else str(roofline_flop_rate[dtype])
        )

        perf_table.append([dtype.name, measured_flops, roofline_flops])

    return tabulate(perf_table, tablefmt="fancy_grid")


def stringify_comparison_vs_roofline(
    expr: BatchedEinsum,
    *,
    schedule: ContractionSchedule | None = None,
    transform: TransformT,
    cl_ctx: cl.Context,
    long_dim_length: int = 100000,
    ignore_unknown_device: bool = False,
) -> str:
    """
    Returns the prettified comparison of *expr* transformed with *transform*
    wrt roofline. The roofline model assumes that kernel's performance is
    limited by either the device's global memory bandwidth or the device's
    floating point units being saturated to their maximum throughput.

    :param ignore_unknown_device: If *False* raises an error if a roofline
        model is unknown for the device in *cl_ctx*. If *True*, no error is
        raised and the roofline performance is marked as "N/A" in the output.
    """

    (dev,) = cl_ctx.devices

    measured_flop_rate = measure_giga_op_rate(
        expr,
        transform=transform,
        schedule=schedule,
        cl_ctx=cl_ctx,
        long_dim_length=long_dim_length,
    )

    try:
        roofline_flop_rate = get_roofline_flop_rate(expr, dev.name)
    except NoDevicePeaksInfoError:
        return _strify_measured_vs_roofline(
            measured_flop_rate, dict.fromkeys(measured_flop_rate.keys(), "N/A")
        )
    else:
        return _strify_measured_vs_roofline(measured_flop_rate, roofline_flop_rate)



import logging

import loopy as lp
import numpy as np
import pyopencl as cl

import feinsum as fnsm

logging.basicConfig(level="INFO")
logger = logging.getLogger(__name__)


def transform_div(t_unit, insn_match=None, kernel_name=None):
    ndim = 3
    ndofs = 35
    ref_einsum = fnsm.batched_einsum(
        "es, sij, ej -> ei",
        [(np.inf, ndim), (ndim, ndofs, ndofs), (np.inf, ndofs)],
        dtypes="float64",
        use_matrix=[
            [{"Jx"}, {"R"}, {"ux"}],
            [{"Jy"}, {"R"}, {"uy"}],
            [{"Jz"}, {"R"}, {"uz"}],
        ],
    )
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    t_unit = lp.tag_inames(t_unit, {subst_map["s"]: "unr"})
    t_unit = lp.split_iname(
        t_unit, subst_map["e"], 8, outer_tag="g.0", inner_tag="l.1"
    )
    t_unit = lp.split_iname(
        t_unit, subst_map["i"], 4, inner_tag="l.0", outer_tag="ilp"
    )

    return t_unit


def transform_grad(t_unit, insn_match=None, kernel_name=None):
    from loopy.match import parse_match

    kernel_name = kernel_name or t_unit.default_entrypoint.name
    insn_match = parse_match(insn_match)

    # {{{ define ref_einsum; get subst_map

    ndim = 3
    ndofs = 35

    ref_einsum = fnsm.einsum(
        "xer,rij,ej->xei",
        fnsm.array(
            (
                ndim,
                np.inf,
                ndim,
            ),
            "float64",
        ),
        fnsm.array((ndim, ndofs, ndofs), "float64"),
        fnsm.array((np.inf, ndofs), "float64"),
        arg_names=["J", "R", "u"],
    )

    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )

    # }}}

    # {{{ transform parameters

    ncells_per_workgroup = 9
    nworkitems_per_cell = 7
    j_tile_len = 9
    i_tile_len = 35

    # }}}

    # {{{ read variables

    i = subst_map["i"]
    j = subst_map["j"]
    e = subst_map["e"]
    R = subst_map["R"]
    u = subst_map["u"]
    x = subst_map["x"]
    r = subst_map["r"]
    out = subst_map["_fe_out"]
    e_inner, e_outer = f"{e}_inner", f"{e}_outer"
    i_inner, i_outer = f"{i}_inner", f"{i}_outer"
    j_inner = f"{j}_inner"
    i_tile, j_tile = f"{i}_tile", f"{j}_tile"
    e_prftch = f"{e}_prftch"
    j_prftch = f"{j}_prftch"
    i_prftch = f"{i}_prftch"
    r_prftch = f"{r}_prftch"
    r_prcmpt = f"{r}_prcmpt"

    # }}}

    # {{{ term hoisting to match the flop count of opt_einsum

    from loopy.symbolic import get_dependencies
    from pymbolic import variables

    knl = t_unit[kernel_name]

    knl = lp.split_reduction_inward(knl, j)
    knl = lp.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, j)

    knl = lp.extract_multiplicative_terms_in_sum_reduction_as_subst(
        knl,
        within=insn_match,
        subst_name="subst",
        arguments=variables(f"{r} {e} {i}"),
        terms_filter=lambda x: (get_dependencies(x) & knl.all_inames()) <= {r, e, i},
    )

    t_unit = t_unit.with_kernel(knl)

    # }}}

    t_unit = lp.split_iname(t_unit, i, i_tile_len, outer_iname=i_tile)
    t_unit = lp.split_iname(t_unit, j, j_tile_len, outer_iname=j_tile)

    t_unit = lp.rename_iname(t_unit, i_inner, i)
    t_unit = lp.rename_iname(t_unit, j_inner, j)
    t_unit = lp.split_iname(
        t_unit, e, ncells_per_workgroup, outer_tag="g.0", inner_tag="l.1"
    )
    t_unit = lp.split_iname(t_unit, i, nworkitems_per_cell, inner_tag="l.0")

    # {{{ TODO: Make precompute smarter (should be a single precompute call)

    t_unit = lp.precompute(
        t_unit,
        "subst",
        sweep_inames=[r],
        precompute_outer_inames=frozenset(
            {e_inner, e_outer, i_inner, i_outer, i_tile}
        ),
        precompute_inames=r_prcmpt,
        temporary_name="tmp_hoist",
        temporary_address_space=lp.AddressSpace.PRIVATE,
        compute_insn_id="insn_hoist",
    )
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, i_outer, only_var_names=["tmp_hoist"]
    )

    t_unit = lp.duplicate_inames(t_unit, i_outer, "id:insn_hoist", "i_outer_hoist")

    # }}}

    # {{{ Move 'u ' to shared.

    # Prefetch 'u' within the tile
    t_unit = lp.precompute(
        t_unit,
        u,
        sweep_inames=[e_inner, j],
        precompute_outer_inames=frozenset([e_outer, i_tile, j_tile]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        precompute_inames=[e_prftch, j_prftch],
        default_tag=None,
    )
    t_unit = lp.split_iname(
        t_unit, e_prftch, ncells_per_workgroup, inner_tag="l.1", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, j_prftch, nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )

    # }}}

    # {{{ Move 'R' to shared.

    t_unit = lp.precompute(
        t_unit,
        R,
        sweep_inames=[r_prcmpt, i_inner, "i_outer_hoist", j],
        precompute_outer_inames=frozenset([e_outer, i_tile, j_tile]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        precompute_inames=[r_prftch, i_prftch, j_prftch],
        default_tag=None,
        within="id:insn_hoist",
    )
    if 0:
        # This branch improves perf. by 20% but, something in loopy
        # non-deterministically leads to very ugly domains.
        t_unit = lp.join_inames(t_unit, [r_prftch, i_prftch, j_prftch], "i_Rprftch")

        t_unit = lp.split_iname(
            t_unit,
            "i_Rprftch",
            ncells_per_workgroup * nworkitems_per_cell,
            outer_tag="unr",
        )

        t_unit = lp.split_iname(
            t_unit,
            "i_Rprftch_inner",
            nworkitems_per_cell,
            inner_tag="l.0",
            outer_tag="l.1",
        )
    else:
        t_unit = lp.split_iname(
            t_unit, i_prftch, ncells_per_workgroup, inner_tag="l.1", outer_tag="unr"
        )
        t_unit = lp.split_iname(
            t_unit, j_prftch, nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
        )

    # }}}

    # {{{ make buffer array smarter (should be a single call to buffer_array)

    t_unit = lp.buffer_array(
        t_unit,
        out,
        buffer_inames=[x],
        init_expression="0",
        default_tag=None,
        temporary_scope=lp.AddressSpace.PRIVATE,
    )
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, i_outer, only_var_names={f"{out}_buf"}
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames=[i_outer],
        within=f"id:init_{out}",
        new_inames=[f"{out}_init_1"],
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames=[i_outer],
        within=f"id:store_{out}",
        new_inames=[f"{out}_store_1"],
    )

    # }}}

    # {{{ must be smarter way of doing this in loopy

    t_unit = lp.realize_reduction(t_unit, insn_id_filter="insn_hoist")
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit,
        frozenset([r_prcmpt, "i_outer_hoist"]),
        only_var_names={f"acc_{j_tile}_{j}"},
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["i_outer_hoist", r_prcmpt],
        within=f"id:insn_hoist_{j_tile}_{j}_init",
        new_inames=["i_outer_hoist_init", f"{r_prcmpt}_init"],
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        ["i_outer_hoist", r_prcmpt],
        within="id:insn_hoist",
        new_inames=["i_outer_hoist_store", f"{r_prcmpt}_store"],
    )

    # }}}

    return t_unit


def transform_face_mass(t_unit, insn_match=None, kernel_name=None):
    kernel_name = kernel_name or t_unit.default_entrypoint.name

    # {{{ define ref_einsum; get subst_map

    nvoldofs = 35
    nfacedofs = 15
    nface = 4
    ref_einsum = fnsm.batched_einsum(
        "ef, fij, fej -> ei",
        [(np.inf, nface), (nface, nvoldofs, nfacedofs), (nface, np.inf, nfacedofs)],
        dtypes="float64",
        use_matrix=[
            [{"J"}, {"R"}, {"v0"}],
            [{"J"}, {"R"}, {"v1"}],
            [{"J"}, {"R"}, {"v2"}],
            [{"J"}, {"R"}, {"v3"}],
        ],
    )
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )

    # }}}

    # {{{ read variables

    e = subst_map["e"]
    e_inner, e_outer = f"{e}_inner", f"{e}_outer"
    f = subst_map["f"]
    i = subst_map["i"]
    j = subst_map["j"]
    # J = subst_map["J"]
    R = subst_map["R"]
    i_0, i_1 = f"{i}_0", f"{i}_1"
    j_0, j_1 = f"{j}_0", f"{j}_1"
    f_0, f_1 = f"{f}_0", f"{f}_1"
    out0, out1, out2, out3 = (
        subst_map["_fe_out"],
        subst_map["_fe_out_0"],
        subst_map["_fe_out_1"],
        subst_map["_fe_out_2"],
    )
    outputs = [out0, out1, out2, out3]
    # v = [subst_map[f"v{i}"] for i in range(4)]

    # }}}

    # {{{ transformation parameters

    ncells_per_group = 16
    nworkitems_per_cell = 12

    # }}}

    from loopy.symbolic import get_dependencies
    from pymbolic import variables

    knl = t_unit[kernel_name]

    for k in range(4):
        knl = lp.extract_multiplicative_terms_in_sum_reduction_as_subst(
            knl,
            subst_name=f"subst{k}",
            arguments=variables(f"{f} {e} {j}"),
            within=f"writes:{outputs[k]}",
            terms_filter=lambda x: get_dependencies(x) & knl.all_inames()
            <= {f, e, j},
        )

    t_unit = t_unit.with_kernel(knl)

    t_unit = lp.split_iname(
        t_unit, e, ncells_per_group, outer_tag="g.0", inner_tag="l.1"
    )

    # {{{ fetch 'R'

    t_unit = lp.precompute(
        t_unit,
        R,
        [f, i, j],
        precompute_outer_inames=frozenset([e_outer]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        default_tag=None,
        precompute_inames=["f_Rprftch", "i_Rprftch", "j_Rprftch"],
    )

    t_unit = lp.split_iname(
        t_unit, "i_Rprftch", ncells_per_group, inner_tag="l.1", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, "j_Rprftch", nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )

    logger.info(f"Prefetched '{R}'")

    # }}}

    t_unit = lp.rename_iname(
        t_unit, i, i_0, within=f"writes:{out0} or writes:{out1}"
    )
    t_unit = lp.rename_iname(
        t_unit, f, f_0, within=f"writes:{out0} or writes:{out1}"
    )
    t_unit = lp.rename_iname(
        t_unit, j, j_0, within=f"writes:{out0} or writes:{out1}"
    )

    t_unit = lp.rename_iname(
        t_unit, i, i_1, within=f"writes:{out2} or writes:{out3}"
    )
    t_unit = lp.rename_iname(
        t_unit, f, f_1, within=f"writes:{out2} or writes:{out3}"
    )
    t_unit = lp.rename_iname(
        t_unit, j, j_1, within=f"writes:{out2} or writes:{out3}"
    )

    t_unit = lp.split_iname(
        t_unit, i_0, nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, i_1, nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )

    for igrp, fgrp, jgrp, prcmpt_tmp, subst in [
        (0, f_0, j_0, "prcmpt_tmp_0", "subst0"),
        (0, f_0, j_0, "prcmpt_tmp_1", "subst1"),
        (1, f_1, j_1, "prcmpt_tmp_0", "subst2"),
        (1, f_1, j_1, "prcmpt_tmp_1", "subst3"),
    ]:
        logger.info(f"Precomputing {subst}")
        t_unit = lp.precompute(
            t_unit,
            subst,
            sweep_inames=[fgrp, e_inner, jgrp],
            precompute_outer_inames=frozenset([e_outer]),
            precompute_inames=[
                f"f_prftch_{igrp}",
                f"e_prftch_{igrp}",
                f"j_prftch_{igrp}",
            ],
            default_tag=None,
            temporary_name=prcmpt_tmp,
            temporary_address_space=lp.AddressSpace.LOCAL,
        )

    for igrp in [0, 1]:
        t_unit = lp.split_iname(
            t_unit,
            f"e_prftch_{igrp}",
            ncells_per_group,
            inner_tag="l.1",
            outer_tag="unr",
        )
        t_unit = lp.split_iname(
            t_unit,
            f"j_prftch_{igrp}",
            nworkitems_per_cell,
            inner_tag="l.0",
            outer_tag="unr",
        )

    t_unit = lp.add_dependency(
        t_unit, "id:subst2 or id:subst3", f"writes:{out0} or writes:{out1}"
    )

    logger.info("Done with transformations.")
    return t_unit


def report_div_performance(cl_ctx):
    ndim = 3
    ndofs = 35
    expr = fnsm.batched_einsum(
        "es, sij, ej -> ei",
        [(np.inf, ndim), (ndim, ndofs, ndofs), (np.inf, ndofs)],
        dtypes="float64",
        use_matrix=[
            [{"Jx"}, {"R"}, {"ux"}],
            [{"Jy"}, {"R"}, {"uy"}],
            [{"Jz"}, {"R"}, {"uz"}],
        ],
    )
    print(
        fnsm.stringify_comparison_vs_roofline(
            expr,
            cl_ctx=cl_ctx,
            transform=transform_div,
            long_dim_length=1000,
            ignore_unknown_device=True,
        )
    )


def report_grad_performance(cl_ctx):
    ndim = 3
    ndofs = 35
    expr = fnsm.einsum(
        "xer,rij,ej->xei",
        fnsm.array(
            (
                ndim,
                np.inf,
                ndim,
            ),
            "float64",
        ),
        fnsm.array((ndim, ndofs, ndofs), "float64"),
        fnsm.array((np.inf, ndofs), "float64"),
        arg_names=["J", "R", "u"],
    )
    print(
        fnsm.stringify_comparison_vs_roofline(
            expr,
            cl_ctx=cl_ctx,
            transform=transform_grad,
            long_dim_length=1000,
            ignore_unknown_device=True,
        )
    )


def report_face_mass_performance(cl_ctx):
    nvoldofs = 35
    nfacedofs = 15
    nface = 4
    expr = fnsm.batched_einsum(
        "ef, fij, fej -> ei",
        [(np.inf, nface), (nface, nvoldofs, nfacedofs), (nface, np.inf, nfacedofs)],
        dtypes="float64",
        use_matrix=[
            [{"J"}, {"R"}, {"v0"}],
            [{"J"}, {"R"}, {"v1"}],
            [{"J"}, {"R"}, {"v2"}],
            [{"J"}, {"R"}, {"v3"}],
        ],
    )

    print(
        fnsm.stringify_comparison_vs_roofline(
            expr,
            cl_ctx=cl_ctx,
            transform=transform_face_mass,
            long_dim_length=1000,
            ignore_unknown_device=True,
        )
    )


def match_and_transfer_tranform():
    t_unit = lp.make_kernel(
        [
            "{[iel_0, idof_0, jdof_0, r_0]:"
            " 0<=iel_0<10000 and 0<=idof_0,jdof_0<35 and 0<=r_0<3}",
            "{[iel_1, idof_1, jdof_1, r_1, x_1]:"
            " 0<=iel_1<10000 and 0<=idof_1,jdof_1<35 and 0<=r_1,x_1<3}",
            "{[iel_2, idof_2, ifacedof, iface]:"
            " 0<=iel_2<10000 and 0<=idof_2<35 and 0<=ifacedof<15 and 0<=iface<4}",
        ],
        """
        # ----- Substitutions
        s_jac_x(_0, _1)       := J[0, _0, _1]
        s_jac_y(_0, _1)       := J[1, _0, _1]
        s_jac_z(_0, _1)       := J[2, _0, _1]
        s_vx(_0, _1)          := vx[_0, _1]
        s_vy(_0, _1)          := vy[_0, _1]
        s_vz(_0, _1)          := vz[_0, _1]
        s_D(_0, _1, _2)       := D[_0, _1, _2]
        s_J(_0, _1, _2)       := J[_0, _1, _2]
        s_u(_0, _1)           := u[_0, _1]
        s_Jface(_0, _1)       := Jface[_0, _1]
        s_Rlift(_0, _1, _2)   := Rlift[_0, _1, _2]
        s_flux_u(_0, _1, _2)  := F_0[_0, _1, _2]
        s_flux_vx(_0, _1, _2) := F_1[_0, _1, _2]
        s_flux_vy(_0, _1, _2) := F_2[_0, _1, _2]
        s_flux_vz(_0, _1, _2) := F_3[_0, _1, _2]
        # ----

        # ----- Div(v)
        div_out_x[iel_0,idof_0] = sum([jdof_0,r_0], \
                                      s_jac_x(iel_0,r_0)*s_D(r_0,idof_0,jdof_0)*s_vx(iel_0,jdof_0))
        div_out_y[iel_0,idof_0] = sum([jdof_0,r_0], \
                                      s_jac_y(iel_0,r_0)*s_D(r_0,idof_0,jdof_0)*s_vy(iel_0,jdof_0))
        div_out_z[iel_0,idof_0] = sum([jdof_0,r_0], \
                                      s_jac_z(iel_0,r_0)*s_D(r_0,idof_0,jdof_0)*s_vz(iel_0,jdof_0))

        ... gbarrier {id=g_barrier_0, dep_query=(writes:div_out_*)}
        # ----- Grad(u)
        with {dep=g_barrier_0}
            grad_out[x_1, iel_1, idof_1] = sum([jdof_1, r_1], \
                                               s_J(x_1, iel_1, r_1)*s_D(r_1, idof_1, jdof_1)*s_u(iel_1, jdof_1))
        end

        ... gbarrier {id=g_barrier_1, dep_query=(writes:grad_out)}
        # ----- Lift(f*)
        with {dep=g_barrier_1}
            lift_0[iel_2, idof_2] = sum([iface, ifacedof], \
                                        s_Jface(iel_2, iface)*s_Rlift(iface, idof_2, ifacedof)*s_flux_u(iface, iel_2, ifacedof))
            lift_1[iel_2, idof_2] = sum([iface, ifacedof], \
                                        s_Jface(iel_2, iface)*s_Rlift(iface, idof_2, ifacedof)*s_flux_vx(iface, iel_2, ifacedof))
            lift_2[iel_2, idof_2] = sum([iface, ifacedof], \
                                        s_Jface(iel_2, iface)*s_Rlift(iface, idof_2, ifacedof)*s_flux_vy(iface, iel_2, ifacedof))
            lift_3[iel_2, idof_2] = sum([iface, ifacedof], \
                                        s_Jface(iel_2, iface)*s_Rlift(iface, idof_2, ifacedof)*s_flux_vz(iface, iel_2, ifacedof))
        end
        """,  # noqa: E501
    )

    t_unit = lp.add_dtypes(
        t_unit,
        {
            arg.name: np.float64
            for arg in t_unit.default_entrypoint.args
            if arg.is_input
        },
    )

    t_unit = transform_div(t_unit, "writes:div_out_*")
    t_unit = transform_grad(t_unit, "writes:grad_out")
    t_unit = transform_face_mass(t_unit, "writes:lift_*")

    print(lp.generate_code_v2(t_unit).device_code())

    return t_unit


if __name__ == "__main__":
    cl_ctx = cl.create_some_context()

    # report_div_performance(cl_ctx)
    # report_grad_performance(cl_ctx)
    # report_face_mass_performance(cl_ctx)
    match_and_transfer_tranform()



import pyopencl as cl

import feinsum as fnsm


def main():
    from feinsum.sql_utils import get_timed_einsums_in_db

    cl_ctx = cl.create_some_context()
    cq = cl.CommandQueue(cl_ctx)
    einsums_in_db = get_timed_einsums_in_db(cq.device)
    print(f"Total einsums in DB = {len(einsums_in_db)}")

    for einsum in einsums_in_db:
        facts = fnsm.query(einsum, cl_ctx)
        print(f"Einsum: {einsum}, Available data: {len(facts)}")


if __name__ == "__main__":
    main()
